import numpy as np
from enum import Enum


class AccessPattern(Enum):
    SEQUENTIAL = 0
    RANDOM = 1


class IOCategory(Enum):
    READ = 1
    WRITE = 2
    METADATA = 3


COL_ACC_PAT = 'acc_pat'
COL_APP_NAME = 'app_name'
COL_BEHAVIOR = 'behavior'
COL_CAT = 'cat'
COL_COUNT = 'count'
COL_FILE_DIR = 'file_dir'
COL_FILE_NAME = 'file_name'
COL_FILE_PATTERN = 'file_pattern'
COL_FUNC_ID = 'func_id'
COL_HOST_NAME = 'host_name'
COL_IO_CAT = 'io_cat'
COL_NODE_NAME = 'node_name'
COL_PROC_NAME = 'proc_name'
COL_RANK = 'rank'
COL_SIZE = 'size'
COL_TIME = 'time'
COL_TIME_OVERALL = 'time_overall'
COL_TIME_RANGE = 'time_range'

VIEW_TYPES = ['time_range', 'file_name', 'proc_name']  # Order matters!
LOGICAL_VIEW_TYPES = [('proc_name', 'app_name'), ('proc_name', 'node_name'), (
    'proc_name', 'rank'), ('file_name', 'file_dir'), ('file_name', 'file_pattern')]

ACC_PAT_SUFFIXES = ['time', 'size', 'count']
DERIVED_MD_OPS = ['close', 'open', 'seek', 'stat']
IO_CATS = [io_cat.value for io_cat in list(IOCategory)]
IO_TYPES = ['read', 'write', 'metadata']
COMPACT_IO_TYPES = ['R', 'W', 'M']

FILE_PATTERN_PLACEHOLDER = '[0-9]'
PROC_NAME_SEPARATOR = '#'

HUMANIZED_VIEW_TYPES = dict(
    app_name='App',
    file_dir='File Directory',
    file_name='File',
    file_pattern='File Pattern',
    node_name='Node',
    proc_name='Process',
    rank='Rank',
    time_range='Time Period',
)

XFER_SIZE_BINS = [
    -np.inf,
    4 * 1024.0,
    16 * 1024.0,
    64 * 1024.0,
    256 * 1024.0,
    1 * 1024.0 * 1024.0,
    4 * 1024.0 * 1024.0,
    16 * 1024.0 * 1024.0,
    64 * 1024.0 * 1024.0,
    np.inf
]
XFER_SIZE_BIN_LABELS = [
    '<4 kiB',
    '4-16 kiB',
    '16-64 kiB',
    '64-256 kiB',
    '256 kiB-1 MiB',
    '1-4 MiB',
    '4-16 MiB',
    '16-64 MiB',
    '>64 MiB',
]
XFER_SIZE_BIN_NAMES = [
    '<4 kiB',
    '4 kiB',
    '16 kiB',
    '64 kiB',
    '256 kiB',
    '1 MiB',
    '4 MiB',
    '16 MiB',
    '64 MiB',
    '>64 MiB'
]

EVENT_ATT_REASONS = 'attach_reasons'
EVENT_COMP_HLM = 'compute_hlm'
EVENT_COMP_MAIN_VIEW = 'compute_main_view'
EVENT_COMP_METBD = 'compute_metric_boundaries'
EVENT_COMP_PERS = 'compute_perspectives'
EVENT_DET_BOT = 'detect_bottlenecks'
EVENT_DET_CHAR = 'detect_characteristics'
EVENT_READ_TRACES = 'read_traces'
EVENT_SAVE_BOT = 'save_bottlenecks'



import logging
import yaml
from dask.distributed import Client, LocalCluster
from dask_jobqueue import LSFCluster
from dataclasses import dataclass
from time import sleep
from typing import List, Literal
from .utils.file_utils import ensure_dir


ClusterType = Literal['local', 'lsf']


@dataclass
class ClusterConfig:
    cluster_type: ClusterType = 'local'
    dashboard_port: int = None
    death_timeout: int = None
    debug: bool = False
    host: str = None
    job_directives_skip: List[str] = None
    job_extra_directives: List[str] = None
    local_dir: str = None
    memory: int = 1600
    n_threads_per_worker: int = 16
    n_workers: int = 8
    processes: bool = True
    use_stdin: bool = True


class ClusterManager(object):

    def __init__(self, working_dir: str, config: ClusterConfig):
        self.config = config
        self.working_dir = working_dir
        ensure_dir(working_dir)
        ensure_dir(f"{working_dir}/worker_logs")

    def boot(self):
        self.cluster = self._initialize_cluster()
        self.client = Client(self.cluster)
        logging.debug("Dashboard address:", self.client.dashboard_link)
        if self.config.cluster_type != 'local':
            self.cluster.scale(self.config.n_workers)
            logging.info(f"Scaling cluster to {self.config.n_workers} nodes")
            self._wait_until_workers_alive()

    def shutdown(self):
        self.client.close()
        self.cluster.close()

    def _initialize_cluster(self):
        dashboard_address = None
        if self.config.dashboard_port is not None:
            assert self.config.host is not None, 'Host address must be specified'
            dashboard_address = f"{self.config.host}:{self.config.dashboard_port}"
        if self.config.cluster_type == 'local':
            return LocalCluster(
                # dashboard_address=dashboard_address,
                host=self.config.host,
                local_directory=self.config.local_dir,
                memory_limit=self.config.memory,
                n_workers=self.config.n_workers,
                processes=self.config.processes,
                silence_logs=logging.DEBUG if self.config.debug else logging.CRITICAL,
            )
        elif self.config.cluster_type == 'lsf':
            return LSFCluster(
                cores=self.config.n_workers * self.config.n_threads_per_worker,
                death_timeout=self.config.death_timeout,
                job_directives_skip=self.config.job_directives_skip,
                job_extra_directives=self.config.job_extra_directives,
                local_directory=self.config.local_dir,
                memory=f"{self.config.memory}GB",
                processes=self.config.n_workers,
                scheduler_options=dict(
                    dashboard_address=dashboard_address,
                    host=self.config.host,
                ),
                use_stdin=self.config.use_stdin,
            )

    def _wait_until_workers_alive(self, sleep_seconds=2):
        current_n_workers = len(self.client.scheduler_info()['workers'])
        while self.client.status == 'running' and current_n_workers < self.config.n_workers:
            current_n_workers = len(self.client.scheduler_info()['workers'])
            logging.debug(
                f"Waiting for workers ({current_n_workers}/{self.config.n_workers})")
            # Try to force cluster to boot workers
            self.cluster._correct_state()
            # Wait
            sleep(sleep_seconds)
        logging.debug('All workers alive')


def load_cluster_config(config_path: str):
    with open(config_path) as config_file:
        config = yaml.safe_load(config_file)
        return ClusterConfig(**config)



from distributed.diagnostics.plugin import SchedulerPlugin
from time import sleep
from tqdm.auto import tqdm


class Counter(SchedulerPlugin):

    def __init__(self):
        self.tasks_fingerprints = {}

    async def start(self, scheduler):
        self.scheduler = scheduler

    def transition(self, key, start, finish, *args, **kwargs):
        ts = self.scheduler.tasks[key]
        if finish in ['memory', 'released']:
            if ts.prefix_key not in self.tasks_fingerprints:
                self.tasks_fingerprints[ts.prefix_key] = {
                    'sum_dur': 0,
                    'n_tasks': 0
                }
            self.tasks_fingerprints[ts.prefix_key] = {
                'sum_dur': self.tasks_fingerprints[ts.prefix_key]['sum_dur'] + self.scheduler.get_task_duration(ts),
                'n_tasks': self.tasks_fingerprints[ts.prefix_key]['n_tasks'] + 1
            }


class ProgressBarPlugin(SchedulerPlugin):
    def __init__(self):
        self.tqdm_map = {}
        self.ts_count = {}
        self.ts_done = {}

    def start(self, scheduler):
        self.scheduler = scheduler

    def initialize(self, futures):
        self.tqdm_map = {}
        dask_graph = futures.__dask_graph__()
        self.ts_count = {}
        self.ts_done = {}
        for key, value in dask_graph.layers.items():
            node = key.split("-")[0]
            if node not in self.ts_count:
                self.ts_count[node] = 0
            self.ts_count[node] = self.ts_count[node] + 1
        for key, count in self.ts_count.items():
            if key not in self.tqdm_map:
                self.tqdm_map[key] = tqdm(total=self.ts_count[key], desc=f"{key}")
                self.ts_done[key] = 0

    def transition(self, key, start, finish, *args, **kwargs):
        ts = self.scheduler.tasks[key]
        if finish in ["memory", "released"]:
            self.tqdm_map[ts.prefix_key].update(1)
            self.ts_done[ts.prefix_key] = ts.prefix_key + 1

    def finalize(self):
        for key, count in self.ts_count.items():
            self.tqdm_map[key].update(self.ts_count[key] - self.ts_done[key])
            sleep(0.1)
            self.tqdm_map[key].close()
        self.tqdm_map = {}
        self.ts_count = {}
        self.ts_done = {}

    def restart(self, scheduler):
        self.tqdm_map = {}



import dask.dataframe as dd
from typing import Dict, List

from .analysis import set_metric_scores
from .types import (
    Metric,
    ScoringPerViewPerMetric,
    ScoringResult,
    ViewKey,
    ViewResult,
    ViewResultsPerViewPerMetric,
)


SCORING_ORDER = dict(
    app_name=('app_name', 'time_range', 'file_name'),
    file_dir=('file_dir', 'proc_name', 'time_range'),
    file_name=('file_name', 'proc_name', 'time_range'),
    file_pattern=('file_pattern', 'proc_name', 'time_range'),
    node_name=('node_name', 'time_range', 'file_name'),
    proc_name=('proc_name', 'time_range', 'file_name'),
    rank=('rank', 'time_range', 'file_name'),
    time_range=('time_range', 'proc_name', 'file_name'),
)


class ViewEvaluator(object):

    def evaluate_views(
        self,
        view_results: ViewResultsPerViewPerMetric,
        metrics: List[Metric],
        metric_boundaries: Dict[Metric, dd.core.Scalar],
    ) -> ScoringPerViewPerMetric:
        # Keep evaluated views
        evaluated_views = {}
        # Run through views for each metric
        for metric in metrics:
            evaluated_views[metric] = {}
            for view_key, view_result in view_results[metric].items():
                # Generate evaluated views
                evaluated_views[metric][view_key] = self._generate_evaluated_views(
                    metric=metric,
                    metric_boundary=metric_boundaries[metric],
                    view_key=view_key,
                    view_result=view_result,
                )
        # Return bottleneck views
        return evaluated_views

    def _generate_evaluated_views(
        self,
        view_key: ViewKey,
        view_result: ViewResult,
        metric: str,
        metric_boundary: dd.core.Scalar,
    ):
        # Get view type
        view_type = view_key[-1]

        records_index = view_result.records.index.persist()

        scored_view = view_result.critical_view \
            .map_partitions(set_metric_scores, view_type=view_type, metric=metric, metric_boundary=metric_boundary) \
            .sort_values(f"{metric}_slope", ascending=True) \
            .persist()

        return ScoringResult(
            critical_view=view_result.critical_view,
            records_index=records_index,
            scored_view=scored_view,
        )



import argparse
import yaml
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from time import time
from typing import List, Literal

from .analyzer_result import AnalysisResult
from .cluster_management import ClusterConfig
from .darshan import DarshanAnalyzer
from .recorder import RecorderAnalyzer
from .types import Metric, OutputType, ViewType

AnalyzerType = Literal['darshan', 'dlp', 'recorder']


@dataclass
class OutputConfig:
    compact: bool = False
    group_behavior: bool = True
    max_bottlenecks: int = 3
    name: str = ''
    root_only: bool = False
    run_db_path: str = None
    show_debug: bool = False
    show_characteristics: bool = True
    show_header: bool = True
    type: OutputType = 'console'
    view_names: List[str] = field(default_factory=list)


@dataclass
class Config:
    type: AnalyzerType
    trace_path: str
    bottleneck_dir: str = None
    checkpoint: bool = True
    checkpoint_dir: str = None
    cluster: ClusterConfig = None
    debug: bool = False
    exclude_bottlenecks: List[str] = field(default_factory=list)
    exclude_characteristics: List[str] = field(default_factory=list)
    logical_view_types: bool = False
    metrics: List[Metric] = field(default_factory=list)
    output: OutputConfig = None
    slope_threshold: int = 45
    time_granularity: int = 1e7  # Recorder's time granularity
    verbose: bool = False
    view_types: List[ViewType] = field(default_factory=list)
    working_dir: str = '.wisio'

    def __post_init__(self):
        if isinstance(self.cluster, dict):
            self.cluster = ClusterConfig(**self.cluster)
        if isinstance(self.output, dict):
            self.output = OutputConfig(**self.output)


def _handle_output(result: AnalysisResult, config: Config):
    output_config = config.output
    if output_config.type == 'console':
        result.output.console(
            compact=output_config.compact,
            group_behavior=output_config.group_behavior,
            max_bottlenecks=output_config.max_bottlenecks,
            name=output_config.name,
            root_only=output_config.root_only,
            show_debug=output_config.show_debug,
            show_characteristics=output_config.show_characteristics,
            show_header=output_config.show_header,
            view_names=output_config.view_names,
        )
    elif output_config.type == 'csv':
        result.output.csv(
            max_bottlenecks=output_config.max_bottlenecks,
            name=output_config.name,
            show_debug=output_config.show_debug,
        )
    elif output_config.type == 'sqlite':
        result.output.sqlite(
            name=output_config.name,
            run_db_path=output_config.run_db_path,
        )


def _load_config(config_path: str):
    with open(config_path) as config_file:
        config = yaml.safe_load(config_file)
        return Config(**config)


def handle_bottleneck(bot_parser: ArgumentParser, args: Namespace):
    if not args.bot_command:
        bot_parser.print_help()
    elif args.bot_command == 'inspect':
        pass


def handle_config(config_parser: ArgumentParser, args: Namespace):
    if not args.config_command:
        config_parser.print_help()
    elif args.config_command == 'create':
        pass


def handle_darshan(analyze_parser: ArgumentParser, config: Config):
    analyzer = DarshanAnalyzer(
        bottleneck_dir=config.bottleneck_dir,
        checkpoint=config.checkpoint,
        checkpoint_dir=config.checkpoint_dir,
        cluster_config=config.cluster,
        debug=config.debug,
        verbose=config.verbose,
        working_dir=config.working_dir,
    )
    result = analyzer.analyze_dxt(
        exclude_bottlenecks=config.exclude_bottlenecks,
        exclude_characteristics=config.exclude_characteristics,
        logical_view_types=config.logical_view_types,
        metrics=config.metrics,
        slope_threshold=config.slope_threshold,
        time_granularity=config.time_granularity,
        trace_path_pattern=config.trace_path,
        view_types=config.view_types,
    )
    _handle_output(config=config, result=result)


def handle_recorder(analyze_parser: ArgumentParser, config: Config):
    analyzer = RecorderAnalyzer(
        bottleneck_dir=config.bottleneck_dir,
        checkpoint=config.checkpoint,
        checkpoint_dir=config.checkpoint_dir,
        cluster_config=config.cluster,
        debug=config.debug,
        verbose=config.verbose,
        working_dir=config.working_dir,
    )
    result = analyzer.analyze_parquet(
        exclude_bottlenecks=config.exclude_bottlenecks,
        exclude_characteristics=config.exclude_characteristics,
        logical_view_types=config.logical_view_types,
        metrics=config.metrics,
        slope_threshold=config.slope_threshold,
        time_granularity=config.time_granularity,
        trace_path=config.trace_path,
        view_types=config.view_types,
    )
    _handle_output(config=config, result=result)


def main():
    parser = argparse.ArgumentParser(description='WisIO')
    subparsers = parser.add_subparsers(title='commands', dest='command')

    base_parser = argparse.ArgumentParser(add_help=False)
    base_parser.add_argument('-c', '--config', required=True, help='Config path')

    analyze_parser = subparsers.add_parser('analyze', help='Analyze trace data', parents=[base_parser])

    bot_parser = subparsers.add_parser('bottleneck', help='Bottleneck inspection')
    bot_parsers = bot_parser.add_subparsers(title='commands', dest='bot_command')
    bot_inspect_parser = bot_parsers.add_parser('inspect', help='Inspect a bottleneck', parents=[base_parser])
    bot_inspect_parser.add_argument('id', help='Bottleneck ID')

    config_parser = subparsers.add_parser('config', help='Config helper')
    config_parsers = config_parser.add_subparsers(title='commands', dest='config_command')
    config_parsers.add_parser('create', help='Create a config file')

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
    elif args.command == 'analyze':
        config = _load_config(args.config)
        if config.type == 'darshan':
            handle_darshan(analyze_parser, config)
        elif config.type == 'recorder':
            handle_recorder(analyze_parser, config)
        else:
            print(f"Invalid analyzer type: {config.type}")
    elif args.command == 'bottleneck':
        handle_bottleneck(bot_parser, args)
    elif args.command == 'config':
        handle_config(config_parser, args)
    else:
        print('Invalid command')


if __name__ == '__main__':
    main()



import darshan
import pandas as pd
from datetime import datetime
from ..constants import PROC_NAME_SEPARATOR, IOCategory


def create_dxt_dataframe(trace_path: str, time_granularity=1e3):

    rep = darshan.DarshanReport(trace_path, read_all=True)

    job = rep.metadata['job']
    if 'start_time' in job:
        job_time = job['end_time'] - job['start_time']
    else:
        job_time = job['end_time_sec'] - job['start_time_sec']

    df = rep.records['DXT_POSIX'].to_df()

    dxt_posix = pd.DataFrame(df)

    read_id = []
    read_rank = []
    read_length = []
    read_offsets = []
    read_end_time = []
    read_start_time = []
    read_operation = []
    read_hostname = []
    read_io_cat = []

    write_id = []
    write_rank = []
    write_length = []
    write_offsets = []
    write_end_time = []
    write_start_time = []
    write_operation = []
    write_hostname = []
    write_io_cat = []

    # start = time.time()
    for r in zip(dxt_posix['rank'], dxt_posix['read_segments'], dxt_posix['write_segments'], dxt_posix['id'], dxt_posix['hostname']):
        if not r[1].empty:
            read_id.append([r[3]] * len((r[1]['length'].to_list())))
            read_rank.append([r[0]] * len((r[1]['length'].to_list())))
            read_length.append(r[1]['length'].to_list())
            read_end_time.append(r[1]['end_time'].to_list())
            read_start_time.append(r[1]['start_time'].to_list())
            read_operation.append(['read'] * len((r[1]['length'].to_list())))
            read_offsets.append(r[1]['offset'].to_list())
            read_hostname.append([r[4]] * len((r[1]['length'].to_list())))
            read_io_cat.append([IOCategory.READ.value]*len((r[1]['length'].to_list())))

        if not r[2].empty:
            write_id.append([r[3]] * len((r[2]['length'].to_list())))     
            write_rank.append([r[0]] * len((r[2]['length'].to_list())))
            write_length.append(r[2]['length'].to_list())
            write_end_time.append(r[2]['end_time'].to_list())
            write_start_time.append(r[2]['start_time'].to_list())
            write_operation.append(['write'] * len((r[2]['length'].to_list())))
            write_offsets.append(r[2]['offset'].to_list())
            write_hostname.append([r[4]] * len((r[2]['length'].to_list())))
            write_io_cat.append([IOCategory.WRITE.value]*len((r[2]['length'].to_list())))

    file_name = [rep.data['name_records'][element] for nestedlist in read_id for element in nestedlist]
    rank = [f"app#localhost#{element}#0" for nestedlist in read_rank for element in nestedlist]
    length = [element for nestedlist in read_length for element in nestedlist]
    offsets = [element for nestedlist in read_offsets for element in nestedlist]
    end_time = [element for nestedlist in read_end_time for element in nestedlist]
    operation = [element for nestedlist in read_operation for element in nestedlist]
    start_time = [element for nestedlist in read_start_time for element in nestedlist]
    time_range = [int(element*time_granularity) for nestedlist in read_start_time for element in nestedlist]
    hostname = [element for nestedlist in read_hostname for element in nestedlist]
    io_cat = [element for nestedlist in read_io_cat for element in nestedlist]

    file_name.extend([rep.data['name_records'][element] for nestedlist in write_id for element in nestedlist])
    rank.extend([f"app#localhost#{element}#0" for nestedlist in write_rank for element in nestedlist])
    length.extend([element for nestedlist in write_length for element in nestedlist])
    offsets.extend([element for nestedlist in write_offsets for element in nestedlist])
    end_time.extend([element for nestedlist in write_end_time for element in nestedlist])
    operation.extend([element for nestedlist in write_operation for element in nestedlist])
    start_time.extend([element for nestedlist in write_start_time for element in nestedlist])
    time_range.extend([int(element*time_granularity) for nestedlist in write_start_time for element in nestedlist])
    hostname.extend([element for nestedlist in write_hostname for element in nestedlist])
    io_cat.extend([element for nestedlist in write_io_cat for element in nestedlist])

    dxt_posix_df = pd.DataFrame(
        {
        'file_name': file_name,
        'proc_name': rank,
        'size': length,
        'end_time': end_time,
        'start_time': start_time,
        'func_id': operation,
        # 'offsets': offsets,
        'hostname': hostname,
        'io_cat': io_cat,
        'time_range': time_range,
        })

    dxt_posix_df['cat'] = 0
    dxt_posix_df['acc_pat'] = 0
    dxt_posix_df['count'] = 1
    dxt_posix_df['time'] = dxt_posix_df['end_time'] - dxt_posix_df['start_time']

    return dxt_posix_df, job_time  


def generate_dxt_records(trace_path: str, time_granularity=1e3):
    def get_dict(row):
        d = {}
        d["size"] = row["length"]
        start_time = row['start_time']
        d["time"] = row["end_time"] - start_time
        d["time_range"] = int(
            (((start_time * 1e6) + (d["time"] * 1e6)/2.0) / time_granularity))
        return d
    report = darshan.DarshanReport(trace_path, read_all=True)
    if "DXT_POSIX" in report.modules:
        t0 = datetime.now()
        df = report.records['DXT_POSIX'].to_df()
        
        print('records',  datetime.now()-t0)
        for val in df.iterrows():
            d = {}

            file_id = val["id"]
            pid = val["rank"]
            tid = 0

            d["hostname"] = val["hostname"]
            d["cat"] = 0  # POSIX
            d["file_name"] = report.data['name_records'][file_id]
            d["proc_name"] = PROC_NAME_SEPARATOR.join(
                ['app', d['hostname'], f"{pid}", f"{tid}"])
            d["acc_pat"] = 0
            d["io_cat"] = IOCategory.METADATA.value

            write_segments = val["write_segments"]
            write_offset = None
            # t0 = datetime.now()
            for _, row in write_segments.iterrows():
                if write_offset is not None and write_offset > row['offset']:
                    d["acc_pat"] = 1
                write_offset = row['offset'] + row['length']
                d["count"] = 1
                d["func_id"] = "write"
                d["io_cat"] = IOCategory.WRITE.value
                d.update(get_dict(row))

                yield dict(**d)
            # print('write_segments', datetime.now() - t0)

            read_segments = val["read_segments"]
            read_offset = None
            # t1 = datetime.now()
            for _, row in read_segments.iterrows():
                if read_offset is not None and read_offset > row['offset']:
                    d["acc_pat"] = 1
                read_offset = row['offset'] + row['length']
                d["count"] = 1
                d["func_id"] = "read"
                d["io_cat"] = IOCategory.READ.value
                d.update(get_dict(row))
                yield dict(**d)
            # print('read_segments', datetime.now() - t1)



import dask.dataframe as dd
import json
import numpy as np
import pandas as pd
from dask.distributed import Future, get_client
from typing import List, Union

from .analyzer import CHECKPOINT_MAIN_VIEW, Analyzer
from .cluster_management import ClusterConfig
from .constants import EVENT_READ_TRACES, IO_CATS
from .types import AnalysisAccuracy, Metric, RawStats, ViewType
from .utils.dask_utils import EventLogger


CAT_POSIX = 0
CHECKPOINT_RAW_STATS = '_raw_stats'
DROPPED_COLS = [
    'app',
    'bandwidth',
    'file_id',
    'hostname',
    'index',
    'level',
    'proc',
    'proc_id',
    'rank',
    'tend',
    'thread_id',
    'tmid',
    'tstart',
]
RENAMED_COLS = {'duration': 'time'}


class RecorderAnalyzer(Analyzer):

    def __init__(
        self,
        working_dir: str,
        bottleneck_dir: str = '',
        checkpoint: bool = False,
        checkpoint_dir: str = '',
        cluster_config: ClusterConfig = None,
        debug=False,
        verbose=False,
    ):
        super().__init__(
            name='Recorder',
            bottleneck_dir=bottleneck_dir,
            checkpoint=checkpoint,
            checkpoint_dir=checkpoint_dir,
            cluster_config=cluster_config,
            debug=debug,
            verbose=verbose,
            working_dir=working_dir,
        )

    def analyze_parquet(
        self,
        trace_path: str,
        metrics: List[Metric],
        accuracy: AnalysisAccuracy = 'pessimistic',
        exclude_bottlenecks: List[str] = [],
        exclude_characteristics: List[str] = [],
        logical_view_types: bool = False,
        slope_threshold: int = 45,
        time_granularity: int = 1e7,
        view_types: List[ViewType] = ['file_name', 'proc_name', 'time_range'],
    ):
        # Init traces
        traces = pd.DataFrame()

        # Check checkpoint status
        main_view_name = self.get_checkpoint_name(CHECKPOINT_MAIN_VIEW, *sorted(view_types))
        if not self.checkpoint or not self.has_checkpoint(name=main_view_name):
            # Read traces
            with EventLogger(key=EVENT_READ_TRACES, message='Read traces'):
                traces, job_time, total_count = self.read_parquet_files(
                    trace_path=trace_path,
                    time_granularity=time_granularity,
                )

        # Prepare raw stats
        raw_stats = self.restore_extra_data(
            name=CHECKPOINT_RAW_STATS,
            fallback=lambda: dict(
                job_time=job_time.persist(),
                time_granularity=time_granularity,
                total_count=total_count.persist(),
            )
        )

        # Analyze traces
        return self.analyze_traces(
            accuracy=accuracy,
            exclude_bottlenecks=exclude_bottlenecks,
            exclude_characteristics=exclude_characteristics,
            logical_view_types=logical_view_types,
            metrics=metrics,
            raw_stats=RawStats(**raw_stats),
            slope_threshold=slope_threshold,
            traces=traces,
            view_types=view_types,
        )

    def compute_job_time(self) -> dd.core.Scalar:
        return self.job_time

    def read_parquet_files(self, trace_path: str, time_granularity: int):
        traces = dd.read_parquet(f"{trace_path}/*.parquet")

        job_time = traces['tend'].max() - traces['tstart'].min()

        traces['acc_pat'] = traces['acc_pat'].astype(np.uint8)
        traces['count'] = 1
        traces['duration'] = traces['duration'].astype(np.float64)
        traces['io_cat'] = traces['io_cat'].astype(np.uint8)

        global_min_max = self._load_global_min_max(trace_path=trace_path)
        time_ranges = self._compute_time_ranges(
            global_min_max=global_min_max,
            time_granularity=time_granularity,
        )

        traces = traces[(traces['cat'] == CAT_POSIX) & (traces['io_cat'].isin(IO_CATS))] \
            .map_partitions(self._set_time_ranges, time_ranges=time_ranges) \
            .rename(columns=RENAMED_COLS) \
            .drop(columns=DROPPED_COLS, errors='ignore')

        total_count = traces.index.count()

        return traces, job_time, total_count

    @staticmethod
    def _compute_time_ranges(global_min_max: dict, time_granularity: int):
        tmid_min, tmid_max = global_min_max['tmid']
        time_ranges = np.arange(tmid_min, tmid_max, time_granularity)
        return get_client().scatter(time_ranges)

    @staticmethod
    def _load_global_min_max(trace_path: str) -> dict:
        with open(f"{trace_path}/global.json") as file:
            global_min_max = json.load(file)
        return global_min_max

    @staticmethod
    def _set_time_ranges(df: pd.DataFrame, time_ranges: Union[Future, np.ndarray]):
        if isinstance(time_ranges, Future):
            time_ranges = time_ranges.result()
        return df.assign(time_range=np.digitize(df['tmid'], bins=time_ranges, right=True))



import logging
import dask.dataframe as dd
from distributed import get_client
from .logger import ElapsedTimeLogger


class EventLogger(ElapsedTimeLogger):

    def __init__(self, key: str, message: str, level=logging.INFO):
        super().__init__(message, level, stacklevel=4)
        self.key = key

    def __enter__(self):
        super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        get_client().log_event('elapsed_times', dict(
            elapsed_time=self.elapsed_time,
            key=self.key,
            message=self.message,
            start_time=self.start_time,
            end_time=self.end_time,
        ))


def flatten_column_names(ddf: dd.DataFrame):
    ddf.columns = ['_'.join(tup).rstrip('_') for tup in ddf.columns.values]
    return ddf



import yaml
from os.path import isfile
from typing import Callable


def load_persisted(path: str, fallback: Callable, persist=True):
    if isfile(path):
        with open(path, 'r') as file:
            return yaml.safe_load(file)
    data = fallback()
    if persist:
        with open(path, 'w') as file:
            yaml.dump(data, file, sort_keys=True)
    return data



import dask.dataframe as dd
import inflect
import math
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sqlite3
import venn
from dask.base import compute
from dataclasses import asdict, dataclass
from distributed import get_client
from matplotlib import ticker
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from pandas.api.types import is_numeric_dtype
from rich.console import Console
from rich.panel import Panel
from rich.padding import Padding
from rich.table import Table
from rich.terminal_theme import DEFAULT_TERMINAL_THEME
from rich.tree import Tree
from scipy.stats import skew
from typing import Dict, List, Set, Tuple

from .analysis import SCORE_BINS, SCORE_INITIALS, SCORE_NAMES
from .constants import (
    COL_APP_NAME,
    COL_FILE_DIR,
    COL_FILE_NAME,
    COL_FILE_PATTERN,
    COL_NODE_NAME,
    COL_PROC_NAME,
    COL_RANK,
    COL_TIME_RANGE,
    EVENT_ATT_REASONS,
    EVENT_COMP_HLM,
    EVENT_COMP_MAIN_VIEW,
    EVENT_COMP_PERS,
    EVENT_DET_BOT,

    EVENT_READ_TRACES,
    EVENT_SAVE_BOT,
)
from .rules import HUMANIZED_KNOWN_RULES, MAX_REASONS, BottleneckRule, KnownCharacteristics
from .types import (
    AnalysisRuntimeConfig,
    BottleneckOutput,
    Characteristics,
    MainView,
    Metric,
    RawStats,
    Score,
    ScoringPerViewPerMetric,
    ViewKey,
    ViewResultsPerViewPerMetric,
    humanized_view_name,
    view_name,
)
from .utils.dask_utils import flatten_column_names
from .utils.file_utils import ensure_dir


@dataclass
class AnalyzerResultOutputCharacteristicsType:
    complexity: float
    io_time: float
    job_time: float
    num_apps: int
    num_files: int
    num_nodes: int
    num_ops: int
    num_procs: int
    num_time_periods: int
    per_io_time: float


@dataclass
class AnalyzerResultOutputCountsType:
    raw_count: int
    hlm_count: int
    main_view_count: int
    avg_perspective_count: Dict[str, int]
    avg_perspective_count_std: Dict[str, float]
    avg_perspective_critical_count: Dict[str, int]
    avg_perspective_critical_count_std: Dict[str, float]
    perspective_skewness: Dict[str, float]
    root_perspective_skewness: Dict[str, float]
    per_records_discarded: Dict[str, float]
    per_records_retained: Dict[str, float]
    num_bottlenecks: Dict[str, int]
    num_metrics: int
    num_perspectives: int
    num_rules: int
    evaluated_records: Dict[str, int]
    perspective_count_tree: Dict[str, Dict[str, int]]
    perspective_critical_count_tree: Dict[str, Dict[str, int]]
    perspective_record_count_tree: Dict[str, Dict[str, int]]
    reasoned_records: Dict[str, int]
    slope_filtered_records: Dict[str, int]


@dataclass
class AnalyzerResultOutputSeveritiesType:
    critical_count: Dict[str, int]
    critical_tree: Dict[str, Dict[str, int]]
    very_high_count: Dict[str, int]
    very_high_tree: Dict[str, Dict[str, int]]
    high_count: Dict[str, int]
    high_tree: Dict[str, Dict[str, int]]
    medium_count: Dict[str, int]
    medium_tree: Dict[str, Dict[str, int]]
    low_count: Dict[str, int]
    very_low_count: Dict[str, int]
    trivial_count: Dict[str, int]
    none_count: Dict[str, int]
    root_critical_count: Dict[str, int]
    root_very_high_count: Dict[str, int]
    root_high_count: Dict[str, int]
    root_medium_count: Dict[str, int]
    root_low_count: Dict[str, int]
    root_very_low_count: Dict[str, int]
    root_trivial_count: Dict[str, int]
    root_none_count: Dict[str, int]


@dataclass
class AnalyzerResultOutputThroughputsType:
    bottlenecks: Dict[str, float]
    evaluated_records: Dict[str, float]
    perspectives: Dict[str, float]
    reasoned_records: Dict[str, float]
    rules: Dict[str, float]
    slope_filtered_records: Dict[str, float]


@dataclass
class AnalyzerResultOutputTimingsType:
    read_traces: Dict[str, float]
    compute_hlm: Dict[str, float]
    compute_main_view: Dict[str, float]
    compute_perspectives: Dict[str, float]
    detect_bottlenecks: Dict[str, float]
    attach_reasons: Dict[str, float]
    save_bottlenecks: Dict[str, float]


@dataclass
class AnalyzerResultOutputType:
    _bottlenecks: List[List[BottleneckOutput]]
    _characteristics: Characteristics
    _raw_stats: RawStats
    characteristics: AnalyzerResultOutputCharacteristicsType
    counts: AnalyzerResultOutputCountsType
    runtime_config: AnalysisRuntimeConfig
    severities: AnalyzerResultOutputSeveritiesType
    throughputs: AnalyzerResultOutputThroughputsType
    timings: AnalyzerResultOutputTimingsType


def _colored_description(description: str, score: str = None):
    if score is None:
        return description
    if score == SCORE_NAMES[0]:
        return description
    elif score == SCORE_NAMES[1]:
        return f"[light_cyan3]{description}"
    elif score == SCORE_NAMES[2]:
        return f"[chartreuse2]{description}"
    elif score == SCORE_NAMES[3]:
        return f"[yellow4]{description}"
    elif score == SCORE_NAMES[4]:
        return f"[yellow3]{description}"
    elif score == SCORE_NAMES[5]:
        return f"[orange3]{description}"
    elif score == SCORE_NAMES[6]:
        return f"[dark_orange3]{description}"
    elif score == SCORE_NAMES[7]:
        return f"[red3]{description}"


def _create_characteristics_table(
    characteristics: Characteristics,
    compact: bool,
    job_time: float,
):
    char_table = Table(box=None, show_header=False)
    char_table.add_column(style="cyan")
    char_table.add_column()

    if compact:
        char_table.add_row('Runtime', f"{job_time:.2f} s")
    else:
        char_table.add_row('Runtime', f"{job_time:.2f} seconds")

    # Add each key-value pair to the table as a row
    for char in characteristics.values():
        if compact or char.detail_list is None or len(char.detail_list) == 0:
            if compact:
                char_table.add_row(char.description, char.compact_desc)
            else:
                char_table.add_row(char.description, char.value_fmt)
        else:
            detail_tree = Tree(char.value_fmt)
            for detail in char.detail_list:
                detail_tree.add(detail)
            char_table.add_row(char.description, detail_tree)

    return char_table


def _create_bottleneck_table(
    bottleneck_rules: Dict[str, BottleneckRule],
    bottlenecks: pd.DataFrame,
    compact: bool,
    max_bottlenecks: int,
    metric: str,
    pluralize: inflect.engine,
):
    view_names = list(bottlenecks['view_name'].unique())

    bot_table = Table(box=None, show_header=False)

    total_bot_count = 0
    total_reason_count = 0

    for view_name in view_names:
        view_bottlenecks = bottlenecks[bottlenecks['view_name'] == view_name]
        bot_count = len(view_bottlenecks)
        total_bot_count = total_bot_count + bot_count
        if bot_count == 0:
            continue
        reason_cols = view_bottlenecks.columns[view_bottlenecks.columns.str.contains('reason')]
        reason_rules = [(reason.split('.')[0], reason) for reason in reason_cols]
        reason_count = sum(
            len(view_bottlenecks[(view_bottlenecks[rule] & view_bottlenecks[reason])])
            for rule, reason in reason_rules
        )
        total_reason_count = total_reason_count + reason_count
        view_key = tuple(view_name.split('.'))
        view_tree = Tree((
            f"{humanized_view_name(view_key, '>').replace('Period', '').strip()} View "
            f"({bot_count} {pluralize.plural_noun('bottleneck', bot_count)} "
            f"with {reason_count} {pluralize.plural_noun('reason', reason_count)})"
        ))
        for _, bottleneck in view_bottlenecks[:max_bottlenecks].iterrows():
            bot_id = getattr(bottleneck, 'id')
            bot_desc = None
            bot_score = getattr(bottleneck, f"{metric}_score")
            reasons = []
            for rule, rule_impl in bottleneck_rules.items():

                view_type = view_key[-1]

                num_files = int(getattr(bottleneck, 'num_file_name', 0))
                num_processes = int(getattr(bottleneck, 'num_proc_name', 0))
                num_time_periods = int(getattr(bottleneck, 'num_time_range', 0))

                if view_type in [COL_FILE_NAME, COL_FILE_DIR, COL_FILE_PATTERN]:
                    num_files = int(getattr(bottleneck, f"num_{view_type}", 0))
                if view_type in [COL_APP_NAME, COL_NODE_NAME, COL_PROC_NAME, COL_RANK]:
                    num_processes = int(getattr(bottleneck, f"num_{view_type}", 0))

                # TODO move to upper level
                if bot_desc is None:
                    bot_desc = rule_impl.describe_bottleneck(
                        compact=compact,
                        metric=getattr(bottleneck, 'metric'),
                        num_files=num_files,
                        num_ops=int(bottleneck['count']),
                        num_processes=num_processes,
                        num_time_periods=num_time_periods,
                        subject=getattr(bottleneck, 'subject'),
                        time=float(getattr(bottleneck, 'time')),
                        time_overall=float(getattr(bottleneck, 'time_overall')),
                        view_type=view_key[-1],
                    )

                humanized_rule = HUMANIZED_KNOWN_RULES[rule]

                # Check if rule applies
                if getattr(bottleneck, rule, False):
                    num_reasons = len(rule_impl.rule.reasons)

                    # Check if any reason is found
                    if any(getattr(bottleneck, f"{rule}.reason.{i}", False) for i in range(num_reasons)):
                        for i in range(num_reasons):
                            # Check if reason applies
                            if getattr(bottleneck, f"{rule}.reason.{i}", False):
                                reason = rule_impl.describe_reason(
                                    bottleneck=dict(bottleneck),
                                    reason_index=i,
                                )
                                reasons.append(f"[{humanized_rule}] {reason}")
                    else:
                        # TODO give details
                        reasons.append(f"[{humanized_rule}] No reason found, investigation needed!")

            nice_bot_desc = f"[{SCORE_INITIALS[bot_score]}{bot_id}] {bot_desc}"
            bot_tree = Tree(_colored_description(nice_bot_desc, bot_score))
            for reason in reasons:
                bot_tree.add(_colored_description(reason, bot_score))
            view_tree.add(bot_tree)

        if max_bottlenecks > 0 and bot_count > max_bottlenecks:
            remaining_count = bot_count - max_bottlenecks
            view_tree.add(f"({remaining_count} more)")

        bot_table.add_row(view_tree)

    return bot_table, total_bot_count, total_reason_count


class AnalyzerResultOutput(object):

    def __init__(
        self,
        bottleneck_dir: str,
        bottleneck_rules: Dict[str, BottleneckRule],
        characteristics: Characteristics,
        evaluated_views: ScoringPerViewPerMetric,
        main_view: MainView,
        raw_stats: RawStats,
        runtime_config: AnalysisRuntimeConfig,
        view_results: ViewResultsPerViewPerMetric,
    ) -> None:
        self.bottleneck_dir = bottleneck_dir
        self.bottleneck_rules = bottleneck_rules
        self.characteristics = characteristics
        self.evaluated_views = evaluated_views
        self.main_view = main_view
        self.pluralize = inflect.engine()
        self.raw_stats = raw_stats
        self.runtime_config = runtime_config
        self.view_results = view_results

    def console(
        self,
        compact=False,
        group_behavior=True,
        max_bottlenecks=3,
        name='',
        root_only=False,
        show_debug=False,
        show_characteristics=True,
        show_header=True,
        view_names: List[str] = [],
    ):

        # TODO metric
        metric = 'iops'

        output = self._create_output_type(group_behavior=group_behavior)

        bottlenecks = output._bottlenecks
        characteristics = output._characteristics
        raw_stats = output._raw_stats

        if len(view_names) > 0:
            bottlenecks = bottlenecks.query('view_name in @view_names', local_dict={'view_names': view_names})
        elif root_only:
            bottlenecks = bottlenecks[bottlenecks['view_depth'] == 1]

        char_table = _create_characteristics_table(
            characteristics=characteristics,
            compact=compact,
            job_time=output.characteristics.job_time,
        )

        char_panel = Panel(
            renderable=char_table,
            title=' '.join([name, 'I/O Characteristics']).strip() if show_header else None,
            subtitle=(
                '[bold]R[/bold]: Read - '
                '[bold]W[/bold]: Write - '
                '[bold]M[/bold]: Metadata '
            ),
            subtitle_align='left',
            padding=1,
        )

        bot_table, total_bot_count, total_reason_count = _create_bottleneck_table(
            bottleneck_rules=self.bottleneck_rules,
            bottlenecks=bottlenecks,
            compact=compact,
            max_bottlenecks=max_bottlenecks,
            metric=metric,
            pluralize=self.pluralize,
        )

        bot_panel = Panel(
            bot_table,
            title=(
                f"{total_bot_count} I/O {self.pluralize.plural_noun('Bottleneck', total_bot_count)} with "
                f"{total_reason_count} {self.pluralize.plural_noun('Reason', total_reason_count)}"
            ) if show_header else None,
            padding=1
        )

        console = Console(record=True)

        if show_debug:
            main_view_count = output.counts.main_view_count
            raw_total_count = output.counts.raw_count

            debug_table = Table(box=None, show_header=False)
            debug_table.add_column(style="cyan")
            debug_table.add_column()

            retained_tree = Tree(f"raw: {raw_total_count} records (100%)")
            main_view_tree = retained_tree.add(
                f"aggregated view: {main_view_count} ({main_view_count/raw_total_count*100:.2f}% 100%)")

            for metric in output.counts.perspective_count_tree:
                metric_tree = Tree((
                    f"{metric}: "
                    f"{output.counts.avg_perspective_count[metric]:.2f}±{output.counts.avg_perspective_count_std[metric]:.2f} "
                    f"({output.counts.avg_perspective_count[metric]/raw_total_count*100:.2f}% "
                    f"{output.counts.avg_perspective_count[metric]/main_view_count*100:.2f}%)"
                    " -S> "
                    f"{output.counts.avg_perspective_critical_count[metric]:.2f}±{output.counts.avg_perspective_critical_count_std[metric]:.2f} "
                    f"({output.counts.avg_perspective_critical_count[metric]/raw_total_count*100:.2f}% "
                    f"{output.counts.avg_perspective_critical_count[metric]/main_view_count*100:.2f}%)"
                ))
                for count_key in output.counts.perspective_count_tree[metric]:
                    view_count = output.counts.perspective_count_tree[metric][count_key]
                    view_critical_count = output.counts.perspective_critical_count_tree[metric][count_key]
                    metric_tree.add((
                        f"{count_key}: "
                        f"{view_count} ({view_count/raw_total_count*100:.2f}% {view_count/main_view_count*100:.2f}%)"
                        " -S> "
                        f"{view_critical_count} ({view_critical_count/raw_total_count*100:.2f}% {view_critical_count/main_view_count*100:.2f}%)"
                    ))
                main_view_tree.add(metric_tree)

            debug_table.add_row('Retained Records', retained_tree)

            count_table = Table(box=None, show_header=True)
            count_table.add_column()
            count_table.add_column('Count')
            count_table.add_column('Processed Record Count')
            count_table.add_row(
                'Perspectives',
                f"{output.counts.num_perspectives}",
                f"{output.counts.slope_filtered_records[metric]}",
            )
            count_table.add_row(
                'Bottlenecks',
                f"{output.counts.num_bottlenecks[metric]}",
                f"{output.counts.evaluated_records[metric]}",
            )
            count_table.add_row(
                'Rules',
                f"{output.counts.num_rules}",
                f"{output.counts.reasoned_records[metric]}",
            )

            debug_table.add_row('Counts', count_table)

            tput_table = Table(box=None, show_header=True)
            tput_table.add_column()
            tput_table.add_column('Throughput')
            tput_table.add_column('Record Throughput')
            tput_table.add_row(
                'Perspectives',
                f"{output.throughputs.perspectives[metric]:.2f} perspectives/sec",
                f"{output.throughputs.slope_filtered_records[metric]:.2f} records/sec",
            )
            tput_table.add_row(
                'Bottlenecks',
                f"{output.throughputs.bottlenecks[metric]:.2f} bottlenecks/sec",
                f"{output.throughputs.evaluated_records[metric]:.2f} records/sec",
            )
            tput_table.add_row(
                'Rules',
                f"{output.throughputs.rules[metric]:.2f} rules/sec",
                f"{output.throughputs.reasoned_records[metric]:.2f} records/sec",
            )

            debug_table.add_row('Throughputs', tput_table)

            tot_bottlenecks = 0
            for metric in output.counts.num_bottlenecks:
                num_bottlenecks = output.counts.num_bottlenecks[metric]
                tot_bottlenecks = tot_bottlenecks + num_bottlenecks
            severity_tree = Tree(
                f"total: {tot_bottlenecks} bottlenecks (100%)")
            for metric in output.counts.num_bottlenecks:
                num_bottlenecks = output.counts.num_bottlenecks[metric]
                severity_metric_tree = severity_tree.add((
                    f"{metric}: "
                    f"{num_bottlenecks} "
                    f"({num_bottlenecks/tot_bottlenecks*100:.2f}% 100%)"
                ))
                severity_metric_tree.add((
                    f"critical: "
                    f"{output.severities.critical_count[metric]} "
                    f"({output.severities.critical_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.critical_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"very high: "
                    f"{output.severities.very_high_count[metric]} "
                    f"({output.severities.very_high_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.very_high_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"high: "
                    f"{output.severities.high_count[metric]} "
                    f"({output.severities.high_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.high_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"medium: "
                    f"{output.severities.medium_count[metric]} "
                    f"({output.severities.medium_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.medium_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"low: "
                    f"{output.severities.low_count[metric]} "
                    f"({output.severities.low_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.low_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"very low: "
                    f"{output.severities.very_low_count[metric]} "
                    f"({output.severities.very_low_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.very_low_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"trivial: "
                    f"{output.severities.trivial_count[metric]} "
                    f"({output.severities.trivial_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.trivial_count[metric]/num_bottlenecks*100:.2f}%)"
                ))
                severity_metric_tree.add((
                    f"none: "
                    f"{output.severities.none_count[metric]} "
                    f"({output.severities.none_count[metric]/tot_bottlenecks*100:.2f}% "
                    f"{output.severities.none_count[metric]/num_bottlenecks*100:.2f}%)"
                ))

            debug_table.add_row('Severities', severity_tree)

            setup_table = Table(box=None, show_header=False)
            setup_table.add_column()
            setup_table.add_column()
            setup_table.add_row('Accuracy', output.runtime_config.accuracy)
            setup_table.add_row('Checkpoint', 'enabled' if output.runtime_config.checkpoint else 'disabled')
            setup_table.add_row('Cluster memory', f"{output.runtime_config.memory}")
            setup_table.add_row('Cluster # of workers', f"{output.runtime_config.num_workers}")
            setup_table.add_row('Cluster # of threads per workers',
                                f"{output.runtime_config.num_threads_per_worker}")
            setup_table.add_row('Cluster processes',
                                'enabled' if output.runtime_config.processes else 'disabled')
            setup_table.add_row('Cluster type', output.runtime_config.cluster_type)
            setup_table.add_row('Debug', 'enabled' if output.runtime_config.debug else 'disabled')
            setup_table.add_row('Slope threshold', f"{output.runtime_config.slope_threshold:.2f}")
            debug_table.add_row('Runtime Config', setup_table)

            debug_panel = Panel(debug_table, title='Debug', padding=1)

            if show_characteristics:
                console.print(char_panel, Padding(''), bot_panel, Padding(''), debug_panel)
            else:
                console.print(bot_panel, Padding(''), debug_panel)
        else:
            if show_characteristics:
                console.print(char_panel, Padding(''), bot_panel)
            else:
                console.print(bot_panel)

        run_dir = f"{self.runtime_config.working_dir}/{self.runtime_config.run_id}"

        console.save_html(f"{run_dir}/output.html", clear=False)
        console.save_text(f"{run_dir}/output.txt", clear=False)
        console.save_svg(f"{run_dir}/output.svg", theme=DEFAULT_TERMINAL_THEME, title="WisIO", clear=True)

    def csv(self, name: str, max_bottlenecks_per_view_type=3, show_debug=True):

        run_dir = f"{self.runtime_config.working_dir}/{self.runtime_config.run_id}"
        ensure_dir(run_dir)

        self._create_output_df(name=name).to_csv(f"{run_dir}/run.csv", encoding='utf8')

        # timings_df, timings_raw_df = self._create_timings_df()
        # timings_df.sort_values(['type', 'key']).to_csv(f"{run_dir}/timings.csv", encoding='utf8')
        # timings_df.sort_values(['time_start']).to_csv(f"{run_dir}/timings_ordered.csv", encoding='utf8')
        # timings_raw_df.to_csv(f"{run_dir}/timings_raw.csv", encoding='utf8')

        # TODO
        metric = 'iops'
        view_name = 'time_range'
        view_key = tuple([view_name])
        if view_key in self.view_results[metric]:
            view_result = self.view_results[metric][view_key]
            view_result.view[[
                'count',
                'count_per',
                'time',
                'time_per',
                f"{metric}_slope",
            ]] \
                .compute() \
                .to_csv(f"{run_dir}/slope_{view_name}.csv", encoding='utf8')

    def sqlite(self, name: str, run_db_path: str = None):

        run_dir = f"{self.runtime_config.working_dir}/{self.runtime_config.run_id}"
        ensure_dir(run_dir)

        con = sqlite3.connect(f"{run_dir}/result.db")

        output_df = self._create_output_df(name=name).reset_index()
        output_df['key'] = output_df['type'] + '_' + output_df['value']
        output_df = output_df.drop(columns=['type', 'value'])
        output_sql_df = output_df.set_index('key').T

        if run_db_path is None:
            output_sql_df.to_sql('run', con=con, index_label='key')
        else:
            run_con = sqlite3.connect(run_db_path)
            output_sql_df.to_sql('run', con=run_con, index_label='key', if_exists='append')
            run_con.close()

        # timings_df, timings_raw_df = self._create_timings_df()
        # timings_df.sort_values(['type', 'key']).to_sql('timings', con=con)
        # timings_df.sort_values(['time_start']).to_sql('timings_ordered', con=con)
        # timings_raw_df.to_sql('timings_raw', con=con)

        # TODO
        metric = 'iops'
        view_name = 'time_range'
        view_key = tuple([view_name])
        if view_key in self.view_results[metric]:
            view_result = self.view_results[metric][view_key]
            view_result.view[[
                'count',
                'count_per',
                'time',
                'time_per',
                f"{metric}_slope",
            ]] \
                .compute() \
                .to_sql(f"slope_{view_name}", con=con)

        con.close()

    def _create_output_df(self, name: str):
        output = self._create_output_type()
        output_dict = asdict(obj=output)
        for output_key in output_dict.copy():
            if output_key.startswith('_'):
                output_dict.pop(output_key)
        output_df = pd.DataFrame.from_dict(output_dict, orient='index') \
            .stack() \
            .to_frame()

        dropping_indices = []
        for ix, row in output_df.copy().iterrows():
            if isinstance(row[0], dict):
                dropping_indices.append(ix)
                type_type = ix[0]
                value_type = ix[1]
                nested_df = pd.json_normalize(row[0])
                for _, nested_row in nested_df.iterrows():
                    secondary_ix = None
                    value = None
                    if '_tree' in value_type:
                        for nested_col in nested_df.columns:
                            suffix = value_type.replace('_tree', '')
                            secondary_ix = f"{nested_col.replace('.', '_').replace('>', '_')}_{suffix}"
                            value = nested_row[nested_col]
                            output_df.loc[(type_type, secondary_ix),] = float(value)
                    else:
                        secondary_ix = f"{nested_row.index[0]}__{value_type}"
                        value = nested_row[0]
                        output_df.loc[(type_type, secondary_ix),] = float(value)

        output_df = output_df \
            .drop(index=dropping_indices) \
            .rename(columns={output_df.columns[0]: name})

        output_df.index.set_names(['type', 'value'], inplace=True)

        return output_df.sort_index()

    def _read_bottlenecks(self, metric: str, group_behavior=True):
        bottlenecks = dd.read_parquet(self.bottleneck_dir)

        if group_behavior:

            cols = bottlenecks.columns

            agg_dict = {}
            agg_dict['behavior'] = 'count'  # count behaviors
            for col in cols:
                if 'num_' in col:
                    agg_dict[col] = 'sum'
                elif is_numeric_dtype(bottlenecks[col]):
                    agg_dict[col] = 'mean'
                else:
                    agg_dict[col] = 'first'

            bottlenecks = (
                bottlenecks
                .groupby(['view_name', f"{metric}_score", 'behavior'])
                .agg(agg_dict)
                .compute()
            )

            # TODO find a better way
            reason_cols = cols[cols.str.contains('reason')]
            rule_cols = [reason.split('.')[0] for reason in reason_cols]

            for col in [*reason_cols, *rule_cols]:
                bottlenecks[col] = bottlenecks[col].astype(bool)

            for col in cols:
                if 'count' in col or 'num_' in col:
                    bottlenecks[col] = bottlenecks[col].astype(int)

        else:
            bottlenecks = bottlenecks.compute()

        bottlenecks['id'] = np.arange(len(bottlenecks)) + 1  # set ids

        return bottlenecks

    def _create_output_type(self, group_behavior: bool) -> AnalyzerResultOutputType:

        # todo
        metric = 'iops'

        bottlenecks = self._read_bottlenecks(metric=metric, group_behavior=group_behavior)

        characteristics, = compute(self.characteristics)
        raw_stats, = compute(self.raw_stats)

        complexity = 0
        io_time = 0
        job_time = float(raw_stats.job_time)
        num_apps = 0
        num_files = 0
        num_nodes = 0
        num_ops = 0
        num_procs = 0
        num_time_periods = 0
        for characteristic in characteristics:
            characteristic_value = characteristics[characteristic].value
            if characteristic == KnownCharacteristics.APP_COUNT.value:
                num_apps = int(characteristic_value)
            elif characteristic == KnownCharacteristics.COMPLEXITY.value:
                complexity = float(characteristic_value)
            elif characteristic == KnownCharacteristics.FILE_COUNT.value:
                num_files = int(characteristic_value)
            elif characteristic == KnownCharacteristics.IO_COUNT.value:
                num_ops = int(characteristic_value)
            elif characteristic == KnownCharacteristics.IO_TIME.value:
                io_time = float(characteristic_value)
            elif characteristic == KnownCharacteristics.NODE_COUNT.value:
                num_nodes = int(characteristic_value)
            elif characteristic == KnownCharacteristics.PROC_COUNT.value:
                num_procs = int(characteristic_value)
            elif characteristic == KnownCharacteristics.TIME_PERIOD.value:
                num_time_periods = int(characteristic_value)
        per_io_time = io_time/job_time

        output_characteristics = AnalyzerResultOutputCharacteristicsType(
            complexity=complexity,
            io_time=io_time,
            job_time=job_time,
            num_apps=num_apps,
            num_files=num_files,
            num_nodes=num_nodes,
            num_ops=num_ops,
            num_procs=num_procs,
            num_time_periods=num_time_periods,
            per_io_time=per_io_time,
        )

        main_view_count = len(self.main_view)
        raw_count = int(raw_stats.total_count)

        perspective_count_tree = {}
        perspective_critical_count_tree = {}
        perspective_record_count_tree = {}
        num_metrics = 0
        perspectives = set()
        for metric in self.view_results:
            perspective_count_tree[metric] = {}
            perspective_critical_count_tree[metric] = {}
            perspective_record_count_tree[metric] = {}
            num_metrics = num_metrics + 1
            for view_key, view_result in self.view_results[metric].items():
                count_key = view_name(view_key, '>')
                bot_important_count = view_result.view.reduction(len, sum)
                view_critical_count = view_result.critical_view.reduction(len, sum)
                view_record_count = view_result.records.reduction(len, sum)
                perspective_count_tree[metric][count_key] = bot_important_count
                perspective_critical_count_tree[metric][count_key] = view_critical_count
                perspective_record_count_tree[metric][count_key] = view_record_count
                perspectives.add(view_key)
        num_metrics = num_metrics
        num_perspectives = len(perspectives)
        perspective_count_tree, perspective_critical_count_tree, perspective_record_count_tree, = compute(
            perspective_count_tree,
            perspective_critical_count_tree,
            perspective_record_count_tree,
        )
        root_view_type_counts = {}
        for metric in self.view_results:
            root_view_type_counts[metric] = []
            for view_key, view_result in self.view_results[metric].items():
                if len(view_key) == 1:
                    view_critical_count = perspective_critical_count_tree[metric][count_key]
                    root_view_type_counts[metric].append(view_critical_count)

        avg_perspective_count = {}
        avg_perspective_count_std = {}
        avg_perspective_critical_count = {}
        avg_perspective_critical_count_std = {}
        per_records_discarded = {}
        per_records_retained = {}
        perspective_skewness = {}
        root_perspective_skewness = {}
        for metric in perspective_count_tree:
            perspective_counts = [perspective_count_tree[metric][count_key]
                                  for count_key in perspective_count_tree[metric]]
            perspective_avg = np.average(perspective_counts)
            perspective_std = np.std(perspective_counts)

            perspective_critical_counts = [perspective_critical_count_tree[metric][count_key]
                                           for count_key in perspective_critical_count_tree[metric]]
            perspective_critical_avg = np.average(perspective_critical_counts)
            perspective_critical_std = np.std(perspective_critical_counts)

            avg_perspective_count[metric] = perspective_avg
            avg_perspective_count_std[metric] = perspective_std
            avg_perspective_critical_count[metric] = perspective_critical_avg
            avg_perspective_critical_count_std[metric] = perspective_critical_std

            perspective_critical_per = perspective_critical_avg/raw_count
            per_records_discarded[metric] = 1-perspective_critical_per
            per_records_retained[metric] = perspective_critical_per

            perspective_skewness[metric] = abs(skew(perspective_counts))
            root_perspective_skewness[metric] = abs(skew(root_view_type_counts[metric]))

        num_rules = len(self.bottleneck_rules)
        view_names = list(bottlenecks['view_name'].unique())

        bot_total_count = {}
        bot_critical_count = {}
        bot_very_high_count = {}
        bot_high_count = {}
        bot_medium_count = {}
        bot_low_count = {}
        bot_very_low_count = {}
        bot_trivial_count = {}
        bot_none_count = {}

        bot_total_count[metric] = len(bottlenecks)
        bot_critical_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.CRITICAL.value])
        bot_very_high_count[metric] = len(
            bottlenecks[bottlenecks[f"{metric}_score"] == Score.VERY_HIGH.value])
        bot_high_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.HIGH.value])
        bot_medium_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.MEDIUM.value])
        bot_low_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.LOW.value])
        bot_very_low_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.VERY_LOW.value])
        bot_trivial_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.TRIVIAL.value])
        bot_none_count[metric] = len(bottlenecks[bottlenecks[f"{metric}_score"] == Score.NONE.value])

        bot_root_critical_count = {}
        bot_root_very_high_count = {}
        bot_root_high_count = {}
        bot_root_medium_count = {}
        bot_root_low_count = {}
        bot_root_very_low_count = {}
        bot_root_trivial_count = {}
        bot_root_none_count = {}

        bot_root_critical_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.CRITICAL.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_very_high_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.VERY_HIGH.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_high_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.HIGH.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_medium_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.MEDIUM.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_low_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.LOW.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_very_low_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.VERY_LOW.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_trivial_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.TRIVIAL.value) & (bottlenecks['view_depth'] == 1)])
        bot_root_none_count[metric] = len(
            bottlenecks[(bottlenecks[f"{metric}_score"] == Score.NONE.value) & (bottlenecks['view_depth'] == 1)])

        bot_critical_tree = {}
        bot_very_high_tree = {}
        bot_high_tree = {}
        bot_medium_tree = {}

        bot_critical_tree[metric] = {view_name: len(bottlenecks[(bottlenecks[f"{metric}_score"] == Score.CRITICAL.value) & (
            bottlenecks['view_name'] == view_name)]) for view_name in view_names}
        bot_very_high_tree[metric] = {view_name: len(bottlenecks[(bottlenecks[f"{metric}_score"] == Score.VERY_HIGH.value) & (
            bottlenecks['view_name'] == view_name)]) for view_name in view_names}
        bot_high_tree[metric] = {view_name: len(bottlenecks[(bottlenecks[f"{metric}_score"] == Score.HIGH.value) & (
            bottlenecks['view_name'] == view_name)]) for view_name in view_names}
        bot_medium_tree[metric] = {view_name: len(bottlenecks[(bottlenecks[f"{metric}_score"] == Score.MEDIUM.value) & (
            bottlenecks['view_name'] == view_name)]) for view_name in view_names}

        severities = AnalyzerResultOutputSeveritiesType(
            critical_count=bot_critical_count,
            very_high_count=bot_very_high_count,
            high_count=bot_high_count,
            medium_count=bot_medium_count,
            low_count=bot_low_count,
            very_low_count=bot_very_low_count,
            trivial_count=bot_trivial_count,
            none_count=bot_none_count,
            critical_tree=bot_critical_tree,
            very_high_tree=bot_very_high_tree,
            high_tree=bot_high_tree,
            medium_tree=bot_medium_tree,
            root_critical_count=bot_root_critical_count,
            root_very_high_count=bot_root_very_high_count,
            root_high_count=bot_root_high_count,
            root_medium_count=bot_root_medium_count,
            root_low_count=bot_root_low_count,
            root_very_low_count=bot_root_very_low_count,
            root_trivial_count=bot_root_trivial_count,
            root_none_count=bot_root_none_count,
        )

        elapsed_times = {}
        for _, event in get_client().get_events('elapsed_times'):
            elapsed_times[event['key']] = event['elapsed_time']
        attach_reasons = {}
        compute_hlm = {}
        compute_main_view = {}
        compute_perspectives = {}
        detect_bottlenecks = {}
        read_traces = {}
        save_bottlenecks = {}
        attach_reasons[metric] = float(elapsed_times.get(EVENT_ATT_REASONS, 0))
        compute_hlm[metric] = float(elapsed_times.get(EVENT_COMP_HLM, 0))
        compute_main_view[metric] = float(elapsed_times.get(EVENT_COMP_MAIN_VIEW, 0))
        compute_perspectives[metric] = float(elapsed_times.get(EVENT_COMP_PERS, 0))
        detect_bottlenecks[metric] = float(elapsed_times.get(EVENT_DET_BOT, 0))
        read_traces[metric] = float(elapsed_times.get(EVENT_READ_TRACES, 0))
        save_bottlenecks[metric] = float(elapsed_times.get(EVENT_SAVE_BOT, 0))
        timings = AnalyzerResultOutputTimingsType(
            attach_reasons=attach_reasons,
            compute_hlm=compute_hlm,
            compute_main_view=compute_main_view,
            compute_perspectives=compute_perspectives,
            detect_bottlenecks=detect_bottlenecks,
            read_traces=read_traces,
            save_bottlenecks=save_bottlenecks,
        )

        evaluated_record_dict = {}
        slope_filtered_record_dict = {}
        for metric in self.evaluated_views:
            evaluated_record_dict[metric] = {}
            slope_filtered_record_dict[metric] = {}
            for view_key in self.evaluated_views[metric]:
                scoring = self.evaluated_views[metric][view_key]
                evaluated_record_dict[metric][view_key] = scoring.critical_view.reduction(len, sum)
                slope_filtered_record_dict[metric][view_key] = scoring.records_index.reduction(len, sum)
        evaluated_record_dict, slope_filtered_record_dict, = compute(
            evaluated_record_dict,
            slope_filtered_record_dict,
        )
        evaluated_records = {}
        reasoned_records = {}
        slope_filtered_records = {}
        for metric in self.evaluated_views:
            evaluated_records[metric] = 0
            reasoned_records[metric] = 0
            slope_filtered_records[metric] = 0
            for view_key in self.evaluated_views[metric]:
                view_evaluated_records = evaluated_record_dict[metric][view_key]
                view_slope_filtered_records = slope_filtered_record_dict[metric][view_key]
                evaluated_records[metric] = evaluated_records[metric] + view_evaluated_records
                reasoned_records[metric] = evaluated_records[metric] * num_rules
                slope_filtered_records[metric] = slope_filtered_records[metric] + view_slope_filtered_records

        counts = AnalyzerResultOutputCountsType(
            avg_perspective_count=avg_perspective_count,
            avg_perspective_count_std=avg_perspective_count_std,
            avg_perspective_critical_count=avg_perspective_critical_count,
            avg_perspective_critical_count_std=avg_perspective_critical_count_std,
            evaluated_records=evaluated_records,
            hlm_count=main_view_count,
            main_view_count=main_view_count,
            num_bottlenecks=bot_total_count,
            num_metrics=num_metrics,
            num_perspectives=num_perspectives,
            num_rules=num_rules,
            per_records_discarded=per_records_discarded,
            per_records_retained=per_records_retained,
            perspective_count_tree=perspective_count_tree,
            perspective_critical_count_tree=perspective_critical_count_tree,
            perspective_record_count_tree=perspective_record_count_tree,
            perspective_skewness=perspective_skewness,
            raw_count=raw_count,
            reasoned_records=reasoned_records,
            root_perspective_skewness=root_perspective_skewness,
            slope_filtered_records=slope_filtered_records,
        )

        bottlenecks_tput = {}
        perspectives_tput = {}
        reasoned_records_tput = {}
        rules_tput = {}
        slope_filtered_records_tput = {}
        evaluated_records_tput = {}

        for metric in evaluated_records:
            bottlenecks_tput[metric] = bot_total_count[metric] / \
                detect_bottlenecks[metric]
            evaluated_records_tput[metric] = evaluated_records[metric] / \
                detect_bottlenecks[metric]
            perspectives_tput[metric] = num_perspectives / \
                compute_perspectives[metric]
            reasoned_records_tput[metric] = (
                evaluated_records[metric]*num_rules) / attach_reasons[metric]
            rules_tput[metric] = num_rules / attach_reasons[metric]
            slope_filtered_records_tput[metric] = slope_filtered_records[metric] / \
                compute_perspectives[metric]

        throughputs = AnalyzerResultOutputThroughputsType(
            bottlenecks=bottlenecks_tput,
            evaluated_records=evaluated_records_tput,
            perspectives=perspectives_tput,
            reasoned_records=reasoned_records_tput,
            rules=rules_tput,
            slope_filtered_records=slope_filtered_records_tput,
        )

        return AnalyzerResultOutputType(
            _bottlenecks=bottlenecks,
            _characteristics=characteristics,
            _raw_stats=raw_stats,
            characteristics=output_characteristics,
            counts=counts,
            runtime_config=self.runtime_config,
            severities=severities,
            throughputs=throughputs,
            timings=timings,
        )

    def _create_timings_df(self):
        timing_events = get_client().get_events('timings')

        timings = []
        for _, timing in timing_events:
            timings.append(timing)

        raw_df = pd.DataFrame(timings)

        timings_df = raw_df.copy()
        timings_df['key_type'] = timings_df['key'] + '_' + timings_df['type']

        timings_df = timings_df[['key_type', 'key', 'type', 'time', 'size']] \
            .groupby(['key_type']) \
            .first() \
            .reset_index() \
            .pivot(index='key', columns='type', values=['size', 'time'])

        timings_df['time', 'elapsed'] = timings_df['time', 'end'] - timings_df['time', 'start']

        timings_df = flatten_column_names(timings_df)

        timings_df['type'] = timings_df.index.map(lambda x: x.split('_')[-1])

        return timings_df, raw_df


class AnalysisResultPlots(object):

    def __init__(
        self,
        main_view: MainView,
        view_results: ViewResultsPerViewPerMetric,
        evaluated_views: ScoringPerViewPerMetric,
    ):
        self.main_view = main_view
        self.view_results = view_results
        self.bottlenecks = evaluated_views  # TODO
        self._cmap = plt.cm.get_cmap('RdYlGn')

    def bottleneck_bar(
        self,
        figsize: Tuple[int, int],
        metrics: List[Metric],
        thresholds: List[float],
        markers: List[str],
        colors: List[str],
        labels: List[str] = [],
        marker_size=72,
    ):

        proc_names = list(self.view_results['time'][(COL_PROC_NAME,)].view.index)
        proc_names.sort(key=lambda x: int(x.split('#')[2]))  # order by rank

        dur_data = self.view_results['time'][(COL_PROC_NAME,)].records \
            .groupby([COL_PROC_NAME, COL_TIME_RANGE])['time'] \
            .compute()

        fig, ax = plt.subplots(figsize=figsize)

        bar_data = []
        bar_h = 1

        for y, proc_name in enumerate(proc_names):
            try:
                bar_args = dict(
                    xranges=dur_data.loc[proc_name].to_dict().items(),
                    yrange=(y, bar_h),
                    facecolors='C0',
                    alpha=0.8,
                )

                ax.broken_barh(**bar_args)

                bar_data.append(bar_args)
            except KeyError:
                continue

        scatter_data = {}

        for m, metric in enumerate(metrics):
            scatter_data[metric] = []

            data = self.bottlenecks[metric][(COL_PROC_NAME,)]['mid_level_view'].compute()

            for y, proc_name in enumerate(proc_names):
                try:
                    for time_range, threshold in data.loc[proc_name][f"{metric}_th"].to_dict().items():
                        # print(proc_name, y, time_range, threshold)
                        if threshold >= thresholds[m]:
                            scatter_args = dict(
                                x=time_range,
                                y=y + (bar_h / 2),
                                s=marker_size,
                                c=colors[m],
                                marker=markers[m],
                                alpha=0.6,
                            )
                            ax.scatter(**scatter_args)

                            scatter_data[metric].append(scatter_args)

                except KeyError:
                    continue

        # len(bot_dur_proc_ml.index.get_level_values(0).unique()))
        ax.set_ylim(0, len(proc_names))
        ax.set_xlim(0, max(dur_data.index.get_level_values(1)))
        ax.set_ylabel('Ranks')
        ax.set_xlabel('Job Time')

        legend_handles = [Line2D([0], [0], color='C0', label='I/O Op')]
        for m, metric in enumerate(metrics):
            legend_handles.append(Line2D(
                xdata=[0],
                ydata=[0],
                color='w',
                label=metric if len(labels) == 0 else labels[m],
                marker=markers[m],
                markerfacecolor=colors[m],
                markersize=marker_size / 8,
            ))

        plt.legend(handles=legend_handles, loc='upper right')

        return fig, ax, bar_data, scatter_data

    def bottleneck_timeline(self, metric: Metric):
        return self._bottleneck_timeline_plot(metric=metric, figsize=(10, 5), title=metric)

    def bottleneck_timeline3(
        self,
        metric1: Metric,
        metric2: Metric,
        metric3: Metric,
        figsize: Tuple[int, int],
        label1: str = None,
        label2: str = None,
        label3: str = None,
        threshold=0.0,
        sample_count=0,
    ):
        # plt.style.use('seaborn-poster')
        fig = plt.figure()

        ax1_line, _ = self._bottleneck_timeline_plot(
            metric=metric1,
            figsize=figsize,
            threshold=threshold,
            yaxis_formatter=self._ticker_for_metric(metric1),
            yaxis_label=label1,
            sample_count=sample_count,
            scatter_zorder=4,
        )
        ax1_line.set_xlabel('Timeline')

        ax2 = ax1_line.twinx()
        ax2.spines['right'].set_position(('axes', 1.0))
        self._bottleneck_timeline_plot(
            metric=metric2,
            figsize=figsize,
            ax=ax2,
            color='C4',
            marker='x',
            threshold=threshold,
            yaxis_formatter=self._ticker_for_metric(metric2),
            yaxis_label=label2,
            sample_count=sample_count,
            scatter_zorder=5,
        )

        ax3 = ax1_line.twinx()
        ax3.spines['right'].set_position(('axes', 1.25))
        self._bottleneck_timeline_plot(
            metric=metric3,
            figsize=figsize,
            ax=ax3,
            color='C5',
            marker='v',
            threshold=threshold,
            yaxis_formatter=self._ticker_for_metric(metric3),
            yaxis_label=label3,
            sample_count=sample_count,
            scatter_zorder=6,
        )

        plt.tight_layout()

        legend_handles = [
            Line2D([0], [0], color='C0', label=label1, lw=2, marker='o'),
            Line2D([0], [0], color='C4', label=label2, lw=2, marker='X'),
            Line2D([0], [0], color='C5', label=label3, lw=2, marker='v'),
        ]

        plt.legend(handles=legend_handles, loc='upper right')

        # Add the colorbar separately using the RdYlGn colormap
        # You can adjust the position and size of the colorbar as desired
        cmap = plt.cm.RdYlGn  # Choose the RdYlGn colormap
        norm = plt.Normalize(vmin=0, vmax=1)  # Normalize the data
        mappable = plt.cm.ScalarMappable(norm=norm, cmap=cmap)  # Create the mappable
        mappable.set_array(SCORE_BINS['time'])  # Set the data for the colorbar

        # Position the colorbar within the figure
        # Left, bottom, width, and height are in fractions of the figure size (0 to 1)
        # You can adjust these values
        colorbar_ax = fig.add_axes([0.68, 0.1, 0.2, 0.03])
        colorbar = plt.colorbar(
            mappable, cax=colorbar_ax, orientation='horizontal')
        colorbar.set_ticklabels(['critical', 'medium', 'trivial'])
        # Adjust the font size as needed
        colorbar.ax.tick_params(labelsize=12, pad=2)

        # Add a label to the colorbar
        # colorbar.set_label('Colorbar Label')
        colorbar_label = 'Bottleneck Severity'
        # Position the label at the top of the colorbar
        colorbar.ax.xaxis.set_label_position('top')
        # Adjust font size and labelpad as needed
        colorbar.ax.set_xlabel(colorbar_label, fontsize=12, labelpad=4)

        return fig

    def _bottleneck_timeline_plot(
        self,
        metric: Metric,
        figsize: Tuple[int, int],
        threshold: float = 0,
        ax: Axes = None,
        title: str = None,
        color='C0',
        marker='o',
        marker_size=96,
        yaxis_formatter: ticker.Formatter = None,
        yaxis_label: str = None,
        sample_count=0,
        scatter_zorder=0,
    ):
        bott = self.bottlenecks[metric][(COL_TIME_RANGE,)].bottlenecks
        metric_col = next(col for col in bott.columns if metric in col)
        data = bott.compute()
        ax_line = data[metric_col].plot(
            ax=ax, color=color, figsize=figsize, title=title, alpha=0.8)
        if yaxis_formatter is not None:
            ax_line.yaxis.set_major_formatter(yaxis_formatter)
        if yaxis_label is not None:
            ax_line.yaxis.set_label(yaxis_label)
        filtered_data = data.query(f"{metric}_th >= {threshold}").reset_index()
        if sample_count > 0:
            filtered_data = filtered_data.sort_values(
                f"{metric}_th", ascending=False).head(sample_count)
        colors = np.vectorize(self._color_map)(filtered_data[f"{metric}_th"])
        ax_scatter = filtered_data.plot.scatter(
            ax=ax_line,
            x=COL_TIME_RANGE,
            y=metric_col,
            c=colors,
            cmap=self._cmap,
            marker=marker,
            s=marker_size,
            zorder=scatter_zorder,
        )
        ax_scatter.set_ylabel(yaxis_label)
        return ax_line, ax_scatter

    def metric_relations2(
        self,
        view_key: ViewKey,
        metric1: Metric,
        metric2: Metric,
        label1: str = None,
        label2: str = None,
    ):
        return self._metric_relations(
            view_key=view_key,
            metrics=[metric1, metric2],
            labels=[label1, label2],
        )

    def metric_relations3(
        self,
        view_key: ViewKey,
        metric1: Metric,
        metric2: Metric,
        metric3: Metric,
        label1: str = None,
        label2: str = None,
        label3: str = None,
    ):
        return self._metric_relations(
            view_key=view_key,
            metrics=[metric1, metric2, metric3],
            labels=[label1, label2, label3],
        )

    def _metric_relations(self, view_key: ViewKey, metrics: List[Metric], labels: List[str]):
        sets = [set(self.view_results[metric][view_key].records['id'].unique().compute())
                for metric in metrics]
        labels = self._venn_labels(sets, labels, metrics)
        fig, ax = venn.venn3(labels, names=metrics, figsize=(5, 5))
        # ax.get_legend().remove()
        # fig.tight_layout()
        return fig, ax

    def slope(
        self,
        metric: Metric,
        view_keys: List[ViewKey],
        legends: List[str] = [],
        slope_threshold: int = 45,
        ax: Axes = None,
        xlabel: str = None,
        ylabel: str = None,
        figsize: Tuple[int, int] = None,
        color: str = None,
    ):
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        legend_handles = []
        x_col = f"{metric}_per_rev_cs"
        y_col = 'count_cs_per_rev'
        for i, view_key in enumerate(view_keys):
            view_result = self.view_results[metric][view_key]
            view = view_result.view.compute()
            color = f"C{i}" if color is None else color
            self._plot_slope(
                ax=ax,
                color=color,
                metric=metric,
                slope_threshold=slope_threshold,
                view=view,
                x_col=x_col,
                y_col=y_col,
            )
            if len(legends) > 0:
                legend_handles.append(
                    Line2D([0], [0], color=color, label=legends[i]))
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if len(legend_handles) > 0:
            ax.legend(handles=legend_handles)
        return ax, view

    @staticmethod
    def _plot_slope(
        ax: Axes,
        color: str,
        metric: Metric,
        slope_threshold: int,
        view: pd.DataFrame,
        x_col: str,
        y_col: str,
    ):
        slope_cond = view[f"{metric}_slope"] < slope_threshold
        view.loc[slope_cond, f"{x_col}_line"] = view[x_col]
        line = view[f"{x_col}_line"].to_numpy()
        x = view[x_col].to_numpy()
        y = view[y_col].to_numpy()
        last_non_nan_index = np.where(~np.isnan(line))[0][-1]
        dotted = np.copy(line)
        if np.all(np.isnan(line[last_non_nan_index + 1:])):
            # complete dotted line if all values after last non-nan are nan
            dotted[last_non_nan_index + 1:] = x[last_non_nan_index + 1:]
        mask = np.isfinite(dotted)
        ax.plot(dotted[mask], y[mask], c=color, ls=':')
        ax.plot(line, y, c=color)

    def view_relations2(
        self,
        metric: Metric,
        view_key1: ViewKey,
        view_key2: ViewKey,
        label1: str = None,
        label2: str = None,
    ):
        return self._view_relations(
            metric=metric,
            view_keys=[view_key1, view_key2],
            labels=[label1, label2],
        )

    def view_relations3(
        self,
        metric: Metric,
        view_key1: ViewKey,
        view_key2: ViewKey,
        view_key3: ViewKey,
        label1: str = None,
        label2: str = None,
        label3: str = None,
    ):
        return self._view_relations(
            metric=metric,
            view_keys=[view_key1, view_key2, view_key3],
            labels=[label1, label2, label3],
        )

    def _view_relations(self, metric: Metric, view_keys: List[ViewKey], labels: List[str]):
        names = [view_name(view_key).replace('_', '\_')
                 for view_key in view_keys]
        sets = [set(self.view_results[metric][view_key].records['id'].unique().compute())
                for view_key in view_keys]
        labels = self._venn_labels(sets, labels, names)
        fig, ax = venn.venn3(labels, names=names, figsize=(5, 5))
        # ax.get_legend().remove()
        # fig.tight_layout()
        return fig, ax

    @staticmethod
    def _color_map(threshold: float):
        if threshold >= 0.9:
            return 'red'
        elif threshold >= 0.75:
            return 'darkorange'
        elif threshold >= 0.5:
            return 'orange'
        elif threshold >= 0.25:
            return 'gold'
        elif threshold >= 0.1:
            return 'yellow'
        elif threshold >= 0.01:
            return 'yellowgreen'
        elif threshold >= 0.001:
            return 'limegreen'
        else:
            return 'green'

    def _ticker_for_metric(self, metric: Metric):
        if metric == 'bw':
            return self._ticker_bw_formatter
        elif metric == 'duration':
            return ticker.StrMethodFormatter('{x:.1f}')
        elif metric == 'iops':
            return self._ticker_human_formatter
        return None

    @ticker.FuncFormatter
    def _ticker_human_formatter(x, pos):
        num = float('{:.3g}'.format(x))
        magnitude = 0
        while abs(num) >= 1000:
            magnitude += 1
            num /= 1000.0
        return '{}{}'.format('{:f}'.format(num).rstrip('0').rstrip('.'), ['', 'K', 'M', 'B', 'T'][magnitude])

    @ticker.FuncFormatter
    def _ticker_bw_formatter(x, pos):
        return f"{math.ceil(x / 1024.0 ** 3)}"

    @staticmethod
    def _venn_labels(sets: List[Set], labels: List[str], names: List[str]):
        fixed_labels = {}
        for label, value in venn.get_labels(sets).items():
            label_value = int(value)
            if label_value > 1_000_000:
                fixed_labels[label] = f"{label_value // 1_000_000}M"
            if label_value > 1_000:
                fixed_labels[label] = f"{label_value // 1_000}K"
            else:
                fixed_labels[label] = value

        pos_arr = ['10', '01'] if len(sets) == 2 else ['100', '010', '001']
        for index, pos in enumerate(pos_arr):
            bold_label = names[index] if labels[index] == None else labels[index]
            fixed_labels[pos] = f"{fixed_labels[pos]}\n" + \
                rf"$\bf{{{bold_label}}}$"
        return fixed_labels


class AnalysisResult(object):

    def __init__(
        self,
        bottleneck_dir: str,
        bottleneck_rules: Dict[str, BottleneckRule],
        characteristics: Characteristics,
        evaluated_views: ScoringPerViewPerMetric,
        main_view: MainView,
        metric_boundaries,
        raw_stats: RawStats,
        runtime_config: AnalysisRuntimeConfig,
        view_results: ViewResultsPerViewPerMetric,
    ):
        self.bottleneck_dir = bottleneck_dir
        self.bottleneck_rules = bottleneck_rules
        self.characteristics = characteristics
        self.evaluated_views = evaluated_views
        self.main_view = main_view
        self.metric_boundaries = metric_boundaries
        self.raw_stats = raw_stats
        self.runtime_config = runtime_config
        self.view_results = view_results

        self.output = AnalyzerResultOutput(
            bottleneck_dir=bottleneck_dir,
            bottleneck_rules=bottleneck_rules,
            characteristics=characteristics,
            evaluated_views=evaluated_views,
            main_view=main_view,
            raw_stats=raw_stats,
            runtime_config=runtime_config,
            view_results=view_results,
        )

        self.plots = AnalysisResultPlots(
            evaluated_views=evaluated_views,
            main_view=main_view,
            view_results=view_results,
        )



import json
import logging
import numpy as np
from dask import delayed
from dask.bag import read_text
from glob import glob
from typing import List
from .analyzer import Analyzer
from .cluster_management import ClusterConfig
from .constants import (
    COL_ACC_PAT,
    COL_COUNT,
    COL_FILE_NAME,
    COL_FUNC_ID,
    COL_HOST_NAME,
    COL_IO_CAT,
    COL_PROC_NAME,
    COL_TIME,
    COL_TIME_RANGE,
    EVENT_READ_TRACES,
    IOCategory,
)
from .types import AnalysisAccuracy, RawStats, ViewType
from .utils.dask_utils import EventLogger


CHECKPOINT_RAW_STATS = '_raw_stats'
PFW_COL_MAPPING = {
    'name': COL_FUNC_ID,
    'dur': COL_TIME,
    'hostname': COL_HOST_NAME,
    'trange': COL_TIME_RANGE,
    'filename': COL_FILE_NAME,
}


def _load_objects(line, time_granularity, time_approximate, condition_fn):
    d = {}
    if (
        line is not None
        and line != ""
        and len(line) > 0
        and "[" != line[0]
        and line != "\n"
    ):
        val = {}
        try:
            val = json.loads(line)
            logging.debug(f"Loading dict {val}")
            if "name" in val:
                val["ts"] = int(val["ts"])
                if val["ts"] > 0:
                    d["name"] = val["name"]
                    d["cat"] = val["cat"]
                    d["pid"] = int(val["pid"])
                    d["tid"] = int(val["tid"])
                    val["dur"] = float(val["dur"]) / 1e6
                    val["ts"] = float(val["ts"]) / 1e6
                    d["ts"] = val["ts"]
                    d["dur"] = val["dur"]
                    d["te"] = d["ts"] + d["dur"]
                    # if not time_approximate:
                    #     d["tinterval"] = I.to_string(I.closed(val["ts"] , val["ts"] + val["dur"]))
                    d["trange"] = int(
                        ((val["ts"] + val["dur"]) / 2.0) / time_granularity
                    )
                    d.update(_io_function(val, d, time_approximate, condition_fn))
                    logging.debug(f"built an dictionary for line {d}")
        except ValueError as error:
            logging.error(f"Processing {line} failed with {error}")
    return d


def _get_conditions_default(json_obj):
    io_cond = "POSIX" == json_obj["cat"]
    return False, False, io_cond


def _io_columns(time_approximate=True):
    return {
        'hostname': "string",
        'compute_time': "string" if not time_approximate else np.float64,
        'io_time': "string" if not time_approximate else np.float64,
        'app_io_time': "string" if not time_approximate else np.float64,
        'total_time': "string" if not time_approximate else np.float64,
        'filename': "string",
        'phase': np.int16,
        'size': np.int64,
    }


def _io_function(json_object, current_dict, time_approximate, condition_fn):
    d = {}
    d['app_io_time'] = 0
    d['compute_time'] = 0
    d['filename'] = ''
    d['io_time'] = 0
    d['phase'] = 0
    d['size'] = 0
    d['total_time'] = 0
    if not condition_fn:
        condition_fn = _get_conditions_default
    app_io_cond, compute_cond, io_cond = condition_fn(json_object)
    if time_approximate:
        d["total_time"] = 0
        if compute_cond:
            d["compute_time"] = current_dict["dur"]
            d["total_time"] = current_dict["dur"]
            d["phase"] = 1
        elif io_cond:
            d["io_time"] = current_dict["dur"]
            d["total_time"] = current_dict["dur"]
            d["phase"] = 2
        elif app_io_cond:
            d["total_time"] = current_dict["dur"]
            d["app_io_time"] = current_dict["dur"]
            d["phase"] = 3
    # else:
    #     if compute_cond:
    #         d["compute_time"] = current_dict["tinterval"]
    #         d["total_time"] = current_dict["tinterval"]
    #         d["phase"] = 1
    #     elif io_cond:
    #         d["io_time"] = current_dict["tinterval"]
    #         d["total_time"] = current_dict["tinterval"]
    #         d["phase"] = 2
    #     elif app_io_cond:
    #         d["app_io_time"] = current_dict["tinterval"]
    #         d["total_time"] = current_dict["tinterval"]
    #         d["phase"] = 3
    #     else:
    #         d["total_time"] = I.to_string(I.empty())
    #         d["io_time"] = I.to_string(I.empty())
    if "args" in json_object:
        if "fname" in json_object["args"]:
            d["filename"] = json_object["args"]["fname"]
        if "hostname" in json_object["args"]:
            d["hostname"] = json_object["args"]["hostname"]

        if "POSIX" == json_object["cat"] and "ret" in json_object["args"]:
            if "write" in json_object["name"]:
                d["size"] = int(json_object["args"]["ret"])
            elif "read" in json_object["name"] and "readdir" not in json_object["name"]:
                d["size"] = int(json_object["args"]["ret"])
        else:
            if "image_size" in json_object["args"]:
                d["size"] = int(json_object["args"]["image_size"])
    return d


class DLPAnalyzer(Analyzer):
    def __init__(
        self,
        working_dir: str,
        checkpoint: bool = False,
        checkpoint_dir: str = '',
        cluster_config: ClusterConfig = None,
        debug=False,
        verbose=False,
    ):
        super().__init__(
            name='DLP',
            checkpoint=checkpoint,
            checkpoint_dir=checkpoint_dir,
            cluster_config=cluster_config,
            debug=debug,
            verbose=verbose,
            working_dir=working_dir,
        )

    def analyze_pfw(
        self,
        trace_path_pattern: str,
        accuracy: AnalysisAccuracy = 'pessimistic',
        exclude_bottlenecks: List[str] = [],
        exclude_characteristics: List[str] = [],
        logical_view_types: bool = False,
        metrics=['duration'],
        slope_threshold: int = 45,
        time_granularity: int = 1e6,
        view_types: List[ViewType] = ['file_name', 'proc_name', 'time_range'],
    ):
        # Read traces
        with EventLogger(key=EVENT_READ_TRACES, message='Read traces'):
            traces = self.read_pfw(
                trace_path_pattern=trace_path_pattern,
                time_granularity=time_granularity,
            )

        job_time = traces['te'].max() - traces['ts'].min()

        # Prepare raw stats
        raw_stats = self.restore_extra_data(
            name=CHECKPOINT_RAW_STATS,
            fallback=lambda: dict(
                job_time=delayed(job_time),
                time_granularity=time_granularity,
                total_count=traces.index.count().persist(),
            ),
        )

        # Analyze traces
        return self.analyze_traces(
            accuracy=accuracy,
            exclude_bottlenecks=exclude_bottlenecks,
            exclude_characteristics=exclude_characteristics,
            logical_view_types=logical_view_types,
            metrics=metrics,
            raw_stats=RawStats(**raw_stats),
            slope_threshold=slope_threshold,
            traces=traces,
            view_types=view_types,
        )

    def read_pfw(
        self, trace_path_pattern: str, time_granularity: int, time_approximate=True
    ):
        trace_paths = glob(trace_path_pattern)
        all_files = []
        pfw_pattern = []
        pfw_gz_pattern = []
        for trace_path in trace_paths:
            if trace_path.endswith('.pfw'):
                pfw_pattern.append(trace_path)
                all_files.append(trace_path)
            elif trace_path.endswith('.pfw.gz'):
                pfw_gz_pattern.append(trace_path)
                all_files.append(trace_path)
            else:
                logging.warn(f"Ignoring unsuported file {trace_path}")

        main_bag = None
        if len(pfw_pattern) > 0:
            pfw_bag = (
                read_text(pfw_pattern)
                .map(
                    _load_objects,
                    time_granularity=time_granularity,
                    time_approximate=time_approximate,
                    condition_fn=None,
                )
                .filter(lambda x: "name" in x)
            )
        main_bag = pfw_bag

        if main_bag:
            columns = {
                'name': "string",
                'cat': "string",
                'pid': np.int64,  # 'Int64',
                'tid': np.int64,  # 'Int64',
                'ts': np.float64,  # 'Int64',
                'te': np.float64,  # 'Int64',
                'dur': np.float64,  # 'Int64',
                # 'tinterval': "string" if not time_approximate else np.int64, # 'Int64',
                'trange': np.float64,  # 'Int64'
            }
            columns.update(_io_columns())
            # columns.update(load_cols)
            events = main_bag.to_dataframe(meta=columns).query('ts > 0')
            events = events[~events['name'].isin(['mod_main'])]
            # self.n_partition = math.ceil(total_size.compute() / (128 * 1024 ** 2))
            # logging.debug(f"Number of partitions used are {self.n_partition}")
            # self.events = events.repartition('256MB').persist()
            # _ = wait(self.events)
            events['ts'] = events['ts'] - events['ts'].min()
            events['te'] = events['ts'] + events['dur']
            events['trange'] = events['ts'] // time_granularity
            # self.events = self.events.persist()
            # _ = wait(self.events)

        events[COL_PROC_NAME] = (
            'app#'
            + events['hostname']
            + '#'
            + events['pid'].astype(str)
            + '#'
            + events['tid'].astype(str)
        )

        # ddf[col_name] = ddf[col_name].mask(ddf['func_id'].str.contains(
        # md_op) & ~ddf['func_id'].str.contains('dir'), ddf[col])

        read_cond = 'read'
        write_cond = 'write'
        metadata_cond = 'readlink'

        events[COL_ACC_PAT] = 0
        events[COL_COUNT] = 1
        events[COL_IO_CAT] = 0
        events[COL_IO_CAT] = events[COL_IO_CAT].mask(
            (events['cat'] == 'POSIX')
            & ~events['name'].str.contains(read_cond)
            & ~events['name'].str.contains(write_cond),
            IOCategory.METADATA.value,
        )
        events[COL_IO_CAT] = events[COL_IO_CAT].mask(
            (events['cat'] == 'POSIX')
            & events['name'].str.contains(read_cond)
            & ~events['name'].str.contains(metadata_cond),
            IOCategory.READ.value,
        )
        events[COL_IO_CAT] = events[COL_IO_CAT].mask(
            (events['cat'] == 'POSIX')
            & events['name'].str.contains(write_cond)
            & ~events['name'].str.contains(metadata_cond),
            IOCategory.WRITE.value,
        )

        return events.rename(columns=PFW_COL_MAPPING)



# import darshan as d
import dask.dataframe as dd
import pandas as pd
from dask import delayed
from glob import glob
from typing import List

from ._darshan.analysis import create_dxt_dataframe
from .cluster_management import ClusterConfig
from .constants import EVENT_READ_TRACES
from .analyzer import CHECKPOINT_MAIN_VIEW, Analyzer
from .types import AnalysisAccuracy, Metric, RawStats, ViewType
from .utils.dask_utils import EventLogger


CHECKPOINT_RAW_STATS = '_raw_stats'
DXT_COLS = {
    'acc_pat': "uint64",
    'cat': "string",
    'count': "uint64",
    'file_name': "string",
    'func_id': "string",
    'io_cat': "uint64",
    'proc_name': "string",
    'size': "uint64",
    'time': "float64",
    # 'time': "uint64",
    'time_range': "uint64",
}


class DarshanAnalyzer(Analyzer):

    def __init__(
        self,
        working_dir: str,
        bottleneck_dir: str = '',
        checkpoint: bool = False,
        checkpoint_dir: str = '',
        cluster_config: ClusterConfig = None,
        debug=False,
        verbose=False,
    ):
        super().__init__(
            name='Darshan',
            bottleneck_dir=bottleneck_dir,
            checkpoint=checkpoint,
            checkpoint_dir=checkpoint_dir,
            cluster_config=cluster_config,
            debug=debug,
            verbose=verbose,
            working_dir=working_dir,
        )

    def analyze_dxt(
        self,
        trace_path_pattern: str,
        metrics: List[Metric],
        accuracy: AnalysisAccuracy = 'pessimistic',
        exclude_bottlenecks: List[str] = [],
        exclude_characteristics: List[str] = [],
        logical_view_types: bool = False,
        slope_threshold: int = 45,
        time_granularity: int = 1e3,
        view_types: List[ViewType] = ['file_name', 'proc_name', 'time_range'],
    ):
        # Init empty
        job_time = 0
        traces = dd.from_pandas(data=pd.DataFrame(), npartitions=1)

        # Check checkpoint status
        main_view_name = self.get_checkpoint_name(CHECKPOINT_MAIN_VIEW, *sorted(view_types))
        if not self.checkpoint or not self.has_checkpoint(name=main_view_name):
            # Read traces
            with EventLogger(key=EVENT_READ_TRACES, message='Read traces'):
                traces, job_time = self.read_dxt(
                    trace_path_pattern=trace_path_pattern,
                    time_granularity=time_granularity,
                )

        # Prepare raw stats
        raw_stats = self.restore_extra_data(
            name=CHECKPOINT_RAW_STATS,
            fallback=lambda: dict(
                job_time=delayed(job_time),
                time_granularity=time_granularity,
                total_count=traces.index.count().persist(),
            )
        )

        # Analyze traces
        return self.analyze_traces(
            accuracy=accuracy,
            exclude_bottlenecks=exclude_bottlenecks,
            exclude_characteristics=exclude_characteristics,
            logical_view_types=logical_view_types,
            metrics=metrics,
            raw_stats=RawStats(**raw_stats),
            slope_threshold=slope_threshold,
            traces=traces,
            view_types=view_types,
        )

    def read_dxt(self, trace_path_pattern: str, time_granularity: int):
        trace_paths = glob(trace_path_pattern)

        df = None
        for trace_path in trace_paths:
            if df is None:
                df, job_time = create_dxt_dataframe(trace_path, time_granularity)
            else:
                df2, _ = create_dxt_dataframe(trace_path, time_granularity)
                df = pd.concat([df, df2])

        return dd.from_pandas(df, npartitions=max(1, self.cluster_config.n_workers)), job_time



import dask.dataframe as dd
import numpy as np
import pandas as pd
from typing import Callable, Dict, List, Union
from .types import Metric, Score, ViewType


BW_BINS = [  # bw_ranges = [0, 1, 128, 1024, 1024*64]
    0,  # -- 'critical'
    1024 ** 2,  # 1MB -- 'very high'
    1024 ** 2 * 16,  # 16MB -- 'high',
    1024 ** 2 * 16 * 16,  # 256MB -- 'medium',
    1024 ** 3,  # 1GB -- 'low',
    1024 ** 3 * 16,  # 16GB -- 'very low',
    1024 ** 3 * 16 * 4,  # 64GB -- 'trivial',
    1024 ** 4  # 1TB -- 'none
]
BW_BINS_PER_PROC = [
    1,  # -- 'critical'
    1024 ** 2,  # 1MB -- 'very high'
    1024 ** 2 * 10,  # 10MB -- 'high'
    1024 ** 2 * 128,  # 128MB -- 'medium' --- fast hd
    1024 ** 2 * 256,  # 256MB -- 'low', --- nvme perf
    1024 ** 2 * 512,  # 512MB -- 'very low', --- hbm memory
    1024 ** 3,  # 1GB 'trivial' --- single thread bw for memory
    1024 ** 3 * 64,  # 64GB -- 'none', -- agg bw for memory
]
IS_NORMALIZED: Dict[Metric, bool] = dict(
    att_perf=True,
    bw=True,
    intensity=True,
    iops=True,
    time=False,
)
IS_REVERSED: Dict[Metric, bool] = dict(
    att_perf=True,
    bw=True,
    intensity=False,
    iops=True,
    time=False,
)
SCORE_BINS: Dict[Metric, List[float]] = dict(
    iops=[
        np.tan(np.deg2rad(80)),  # 5.67128182
        np.tan(np.deg2rad(70)),  # 2.74747742
        np.tan(np.deg2rad(60)),  # 1.73205081
        np.tan(np.deg2rad(50)),  # 1.19175359
        np.tan(np.deg2rad(40)),  # 0.83909963
        np.tan(np.deg2rad(30)),  # 0.57735027
        np.tan(np.deg2rad(20)),  # 0.36397023
        np.tan(np.deg2rad(10)),  # 0.17632698
    ],
    time=[
        0,
        0.001,
        0.01,
        0.1,
        0.25,
        0.5,
        0.75,
        0.9
    ]
)
SCORE_INITIALS = {
    'none': 'NA',
    'trivial': 'TR',
    'very low': 'VL',
    'low': 'LO',
    'medium': 'MD',
    'high': 'HI',
    'very high': 'VH',
    'critical': 'CR',
}
SCORE_NAMES = [
    Score.NONE.value,
    Score.TRIVIAL.value,
    Score.VERY_LOW.value,
    Score.LOW.value,
    Score.MEDIUM.value,
    Score.HIGH.value,
    Score.VERY_HIGH.value,
    Score.CRITICAL.value,
]
THRESHOLD_FUNCTIONS: Dict[Metric, Callable[[int], Union[float, int]]] = dict(
    iops=lambda x: np.tan(np.deg2rad(x)),
    time=lambda x: x,
)


def set_bound_columns(ddf: Union[dd.DataFrame, pd.DataFrame], is_initial=False):
    # Min(Peak IOPS, Peak I/O BW x I/O intensity) == higher the better
    # less than 25% of peak attainable performance -- reversed
    if not is_initial:
        ddf['bw_intensity'] = ddf['bw'] * ddf['intensity']
        ddf['att_perf'] = ddf[['iops', 'bw_intensity']].min(axis=1)

    # records less than %10 of attainable BW -- reversed
    ddf['bw'] = ddf['size'] / ddf['time']

    # less than 25% of records -- reversed
    ddf['iops'] = ddf['data_count'] / ddf['time']

    # records which tend towards 1 >> 0.9
    ddf['intensity'] = 0.0
    ddf['intensity'] = ddf['intensity'].mask(ddf['size'] > 0, ddf['count'] / ddf['size'])

    if not is_initial:
        return ddf.drop(columns=['bw_intensity'])
    return ddf


def set_metric_scores(df: pd.DataFrame, view_type: ViewType, metric: Metric, metric_boundary=None):
    bin_col, score_col, slope_col = (
        f"{metric}_bin",
        f"{metric}_score",
        f"{metric}_slope",
    )

    bins = SCORE_BINS[metric]
    names = SCORE_NAMES

    # if IS_NORMALIZED[metric]:
    #     bins = np.multiply(bins, metric_boundary)

    # if IS_REVERSED[metric]:
    #     names = np.flip(names)
    #     thresholds = np.flip(thresholds)

    # if metric == 'bw':
    #     bins = BW_BINS_PER_PROC if view_type == COL_PROC_NAME else BW_BINS

    # if metric in ['bw', 'iops']:
    #     df = df.query(f"{metric} > 0")

    df[bin_col] = np.digitize(df[slope_col], bins=bins, right=True)
    df[score_col] = np.choose(df[bin_col] - 1, choices=names, mode='clip')

    return df


def set_metric_slope(df: pd.DataFrame, metric: Metric):
    slope_col = f"{metric}_slope"

    if metric == 'iops':
        df['time_per'] = df['time'] / df['time'].sum()
        df['count_per'] = df['count'] / df['count'].sum()
        df[slope_col] = df['count_per'] / df['time_per']
    elif metric == 'time':
        df[slope_col] = df[metric] / df[metric].sum()

    # 1. io_time == too trivial -- >50% threshold
    # 2. iops == slope analysis -- <45 degree roc (iops) as the main metric

    # automated bottleneck detection based on optimization function
    # one case is io_time (bin with old bins)
    # second case is iops (bin with roc bins)

    # iops = main metric
    # for the capability of automated bottleneck detection
    # absolute value of io time doesn't give us the rate of change depending on io_time/count

    return df



import abc
import dask.bag as db
import dask.dataframe as dd
import itertools as it
import json
import logging
import math
import os
from dask.base import compute, unpack_collections
from dask.delayed import Delayed
from dask.distributed import fire_and_forget, get_client, wait
from time import time
from typing import Callable, Dict, List, Tuple, Union

from .analysis import THRESHOLD_FUNCTIONS, set_bound_columns, set_metric_slope
from .analysis_utils import set_file_dir, set_file_pattern, set_proc_name_parts
from .analyzer_result import AnalysisResult
from .cluster_management import ClusterConfig, ClusterManager
from .constants import (
    ACC_PAT_SUFFIXES,
    COL_FILE_NAME,
    COL_PROC_NAME,
    DERIVED_MD_OPS,
    EVENT_ATT_REASONS,
    EVENT_COMP_HLM,
    EVENT_COMP_MAIN_VIEW,
    EVENT_COMP_METBD,
    EVENT_COMP_PERS,
    EVENT_DET_BOT,
    EVENT_DET_CHAR,
    EVENT_SAVE_BOT,
    LOGICAL_VIEW_TYPES,
    AccessPattern,
    IOCategory,
)
from .rule_engine import RuleEngine
from .scoring import ViewEvaluator
from .types import (
    AnalysisAccuracy,
    AnalysisRuntimeConfig,
    Metric,
    RawStats,
    ViewKey,
    ViewResult,
    ViewType,
)
from .utils.dask_utils import EventLogger, flatten_column_names
from .utils.file_utils import ensure_dir
from .utils.json_encoders import NpEncoder
from .utils.logger import setup_logging


CHECKPOINT_MAIN_VIEW = '_main_view'
CHECKPOINT_METRIC_BOUNDARIES = '_metric_boundaries'
CHECKPOINT_HLM = '_hlm'
EXTRA_COLS = ['io_cat', 'acc_pat', 'func_id']
HLM_AGG = {
    'time': [sum],
    'count': [sum],
    'size': [min, max, sum],
}
HLM_COLS = {
    'count_sum': 'count',
    'size_sum': 'size',
    'time_sum': 'time',
}
VIEW_AGG = {
    # 'bw': max,
    'count': sum,
    'data_count': sum,
    # 'intensity': max,
    # 'iops': max,
    'size': sum,
    'time': sum,
}
WAIT_ENABLED = True


class Analyzer(abc.ABC):

    run_id: str = None

    def __init__(
        self,
        name: str,
        working_dir: str,
        bottleneck_dir: str = '',
        checkpoint: bool = False,
        checkpoint_dir: str = '',
        cluster_config: ClusterConfig = None,
        debug=False,
        verbose=False,
    ):

        if checkpoint:
            assert checkpoint_dir != '', 'Checkpoint directory must be defined'

        self.bottleneck_dir = bottleneck_dir
        self.checkpoint = checkpoint
        self.checkpoint_dir = checkpoint_dir
        self.cluster_config = cluster_config
        self.debug = debug
        self.name = name
        self.verbose = verbose
        self.working_dir = working_dir

        # Setup logging
        ensure_dir(working_dir)
        setup_logging(filename=f"{working_dir}/{name.lower()}_analyzer.log", debug=debug)
        logging.info(f"Initializing {name} analyzer")

        # Init cluster manager
        self.cluster_manager = ClusterManager(working_dir=working_dir, config=cluster_config)

        # Boot cluster
        self.cluster_manager.boot()

    def analyze_traces(
        self,
        traces: dd.DataFrame,
        metrics: List[Metric],
        raw_stats: RawStats,
        accuracy: AnalysisAccuracy = 'pessimistic',
        exclude_bottlenecks: List[str] = [],
        exclude_characteristics: List[str] = [],
        logical_view_types: bool = False,
        slope_threshold: int = 45,
        view_types: List[ViewType] = ['file_name', 'proc_name', 'time_range'],
    ):
        # Set run id
        self.reset_run_id()

        # Create checkpoint names
        hlm_name = self.get_checkpoint_name(CHECKPOINT_HLM, *sorted(view_types))
        main_view_name = self.get_checkpoint_name(CHECKPOINT_MAIN_VIEW, *sorted(view_types))

        # Check there is a checkpointed main view
        if not self.checkpoint or not self.has_checkpoint(name=main_view_name):
            # Compute high-level metrics
            with EventLogger(key=EVENT_COMP_HLM, message='Compute high-level metrics'):
                hlm = self.restore_view(
                    name=hlm_name,
                    fallback=lambda: self.compute_high_level_metrics(
                        traces=traces,
                        view_types=view_types,
                    ),
                    force=False,
                    persist=True,
                )
                self._wait_all(tasks=hlm)

        # Compute main view
        with EventLogger(key=EVENT_COMP_MAIN_VIEW, message='Compute main view'):
            main_view = self.restore_view(
                name=main_view_name,
                fallback=lambda: self.compute_main_view(
                    hlm=hlm,
                    view_types=view_types,
                ),
                force=False,
                persist=True,
            )
            # TODO remove dropped columns
            main_view = main_view.drop(columns=['bw', 'intensity', 'iops', 'att_perf'], errors='ignore')
            self._wait_all(tasks=main_view)

        # return traces, main_view

        # Compute upper bounds
        with EventLogger(key=EVENT_COMP_METBD, message='Compute metric boundaries'):
            metric_boundaries = self.restore_extra_data(
                name='_'.join([CHECKPOINT_METRIC_BOUNDARIES, *sorted(metrics), *sorted(view_types)]),
                fallback=lambda: self.compute_metric_boundaries(
                    main_view=main_view,
                    metrics=metrics,
                    view_types=view_types,
                ),
            )
            self._wait_all(tasks=metric_boundaries)

        # Compute views
        with EventLogger(key=EVENT_COMP_PERS, message='Compute perspectives'):
            view_results = self.compute_views(
                main_view=main_view,
                metrics=metrics,
                metric_boundaries=metric_boundaries,
                slope_threshold=slope_threshold,
                view_types=view_types,
            )
            if logical_view_types:
                logical_view_results = self.compute_logical_views(
                    main_view=main_view,
                    metrics=metrics,
                    metric_boundaries=metric_boundaries,
                    slope_threshold=slope_threshold,
                    view_types=view_types,
                    view_results=view_results,
                )
                view_results.update(logical_view_results)
            self._wait_all(tasks=view_results)

        # Evaluate views
        view_evaluator = ViewEvaluator()
        with EventLogger(key=EVENT_DET_BOT, message='Detect I/O bottlenecks'):
            evaluated_views = view_evaluator.evaluate_views(
                metric_boundaries=metric_boundaries,
                metrics=metrics,
                view_results=view_results,
            )
            self._wait_all(tasks=evaluated_views)

        # Execute rules
        rule_engine = RuleEngine(rules=[], raw_stats=raw_stats, verbose=self.verbose)
        with EventLogger(key=EVENT_ATT_REASONS, message='Attach reasons to I/O bottlenecks'):
            characteristics = rule_engine.process_characteristics(
                exclude_characteristics=exclude_characteristics,
                main_view=main_view,
                view_results=view_results,
            )
            bottlenecks, bottleneck_rules = rule_engine.process_bottlenecks(
                evaluated_views=evaluated_views,
                exclude_bottlenecks=exclude_bottlenecks,
                metric_boundaries=metric_boundaries,
            )
            self._wait_all(tasks=bottlenecks)

        with EventLogger(key=EVENT_SAVE_BOT, message='Save I/O bottlenecks', level=logging.DEBUG):
            bottleneck_dir = self.save_bottlenecks(bottlenecks=bottlenecks)

        # Return result
        return AnalysisResult(
            bottleneck_dir=bottleneck_dir,
            bottleneck_rules=bottleneck_rules,
            characteristics=characteristics,
            evaluated_views=evaluated_views,
            main_view=main_view,
            metric_boundaries=metric_boundaries,
            raw_stats=raw_stats,
            runtime_config=AnalysisRuntimeConfig(
                accuracy=accuracy,
                checkpoint=self.checkpoint,
                cluster_type=self.cluster_config.cluster_type,
                debug=self.debug,
                memory=self.cluster_config.memory,
                num_threads_per_worker=self.cluster_config.n_threads_per_worker,
                num_workers=self.cluster_config.n_workers,
                processes=self.cluster_config.processes,
                run_id=self.run_id,
                slope_threshold=slope_threshold,
                verbose=self.verbose,
                working_dir=self.working_dir,
            ),
            view_results=view_results,
        )

    def compute_high_level_metrics(
        self,
        traces: dd.DataFrame,
        view_types: list,
        partition_size: str = '256MB',
    ) -> dd.DataFrame:
        # Add `io_cat`, `acc_pat`, and `func_id` to groupby
        groupby = view_types.copy()
        groupby.extend(EXTRA_COLS)
        # Compute high-level metrics
        # hlm = traces \
        #     .groupby(groupby) \
        #     .agg(HLM_AGG, split_out=traces.npartitions) \
        #     .persist() \
        #     .reset_index() \
        #     .repartition(partition_size) \
        #     .persist()
        # hlm = traces \
        #     .groupby(groupby) \
        #     .agg(HLM_AGG, split_out=8) \
        #     .reset_index() \
        #     .persist()
        hlm = traces \
            .groupby(groupby) \
            .agg(HLM_AGG) \
            .persist() \
            .reset_index()
        hlm = flatten_column_names(hlm)
        return hlm.rename(columns=HLM_COLS).persist()

    def compute_main_view(
        self,
        hlm: dd.DataFrame,
        view_types: List[ViewType],
        partition_size: str = '256MB',
    ) -> dd.DataFrame:
        # Set derived columns
        hlm = self._set_derived_columns(ddf=hlm)
        # Compute agg_view
        main_view = hlm \
            .drop(columns=EXTRA_COLS) \
            .groupby(view_types) \
            .sum()
        # main_view = hlm \
        #     .drop(columns=EXTRA_COLS) \
        #     .groupby(view_types) \
        #     .sum() \
        #     .persist() \
        #     .repartition(partition_size)
        # Set hashed ids
        # main_view['id'] = main_view.index.map(set_id)
        main_view['id'] = main_view.index.map(hash)
        # Return main_view
        return main_view.persist()

    def compute_metric_boundaries(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        view_types: List[ViewType],
    ) -> Dict[Metric, dd.core.Scalar]:
        metric_boundaries = {}
        for metric in metrics:
            metric_boundary = None
            if metric == 'iops' or metric == 'time':
                if COL_PROC_NAME in view_types:
                    metric_boundary = main_view \
                        .groupby([COL_PROC_NAME]) \
                        .sum()['time'] \
                        .max() \
                        .persist()
                else:
                    metric_boundary = main_view['time'] \
                        .sum() \
                        .persist()
            elif metric == 'bw':

                pass
            metric_boundaries[metric] = metric_boundary
        return metric_boundaries

    def compute_views(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        metric_boundaries: Dict[Metric, dd.core.Scalar],
        slope_threshold: int,
        view_types: List[ViewType],
    ):
        # Keep view results
        view_results = {}

        # Compute multifaceted views for each metric
        for metric in metrics:

            view_results[metric] = {}

            for view_key in self.view_permutations(view_types=view_types):

                view_type = view_key[-1]
                parent_view_key = view_key[:-1]

                parent_view_result = view_results[metric].get(parent_view_key, None)
                parent_records = main_view if parent_view_result is None else parent_view_result.records

                view_result = self.compute_view(
                    metric=metric,
                    metric_boundary=metric_boundaries[metric],
                    records=parent_records,
                    slope_threshold=slope_threshold,
                    view_key=view_key,
                    view_type=view_type,
                )

                view_results[metric][view_key] = view_result

        return view_results

    def compute_logical_views(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        metric_boundaries: Dict[Metric, dd.core.Scalar],
        slope_threshold: int,
        view_types: List[ViewType],
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ):

        for metric in metrics:

            for view_key in LOGICAL_VIEW_TYPES:

                view_type = view_key[-1]
                parent_view_key = view_key[:-1]
                parent_view_type = parent_view_key[0]

                if parent_view_type not in view_types:
                    continue

                parent_view_result = view_results[metric].get(parent_view_key, None)
                parent_records = main_view if parent_view_result is None else parent_view_result.records

                if view_type not in parent_records.columns:
                    parent_records = self._set_logical_columns(
                        view=parent_records,
                        view_types=[parent_view_type],
                    )

                view_result = self.compute_view(
                    metric=metric,
                    metric_boundary=metric_boundaries[metric],
                    records=parent_records,
                    slope_threshold=slope_threshold,
                    view_key=view_key,
                    view_type=view_type,
                )

                view_results[metric][view_key] = view_result

        return view_results

    def compute_view(
        self,
        records: dd.DataFrame,
        view_key: ViewKey,
        view_type: str,
        metric: str,
        metric_boundary: dd.core.Scalar,
        slope_threshold: int,
    ) -> ViewResult:
        # Compute view
        view = self._compute_view(
            metric=metric,
            metric_boundary=metric_boundary,
            records=records,
            view_type=view_type,
        )

        # Filter by slope
        critical_view = view
        if slope_threshold > 0:
            corrected_threshold = THRESHOLD_FUNCTIONS[metric](slope_threshold)
            critical_view = view \
                .query(f"{metric}_slope <= @threshold", local_dict={'threshold': corrected_threshold}) \
                .persist()

        indices = critical_view.index.unique()

        # Find filtered records
        records = records \
            .query(f"{view_type} in @indices", local_dict={'indices': indices}) \
            .persist()

        # Return views & normalization data
        return ViewResult(
            critical_view=critical_view,
            metric=metric,
            records=records,
            view=view,
            view_type=view_type,
        )

    def get_checkpoint_name(self, *args):
        return '_'.join(args)

    def get_checkpoint_path(self, name: str):
        return f"{self.checkpoint_dir}/{name}"

    def has_checkpoint(self, name: str):
        checkpoint_path = self.get_checkpoint_path(name=name)
        return os.path.exists(f"{checkpoint_path}/_metadata")

    def reset_run_id(self):
        self.run_id = f"{int(time())}"
        ensure_dir(f"{self.working_dir}/{self.run_id}")

    def restore_extra_data(self, name: str, fallback: Callable[[], dict], force=False, persist=False):
        if self.checkpoint:
            data_path = f"{self.get_checkpoint_path(name=name)}.json"
            if force or not os.path.exists(data_path):
                data = fallback()
                fire_and_forget(
                    get_client().submit(
                        self.store_extra_data,
                        data=get_client().submit(compute, data),
                        data_path=data_path
                    )
                )
                return data
            with open(data_path, 'r') as f:
                return json.load(f)
        return fallback()

    def restore_view(self, name: str, fallback: Callable[[], dd.DataFrame], force=False, persist=False):
        if self.checkpoint:
            view_path = self.get_checkpoint_path(name=name)
            if force or not self.has_checkpoint(name=name):
                view = fallback()
                view.to_parquet(f"{view_path}", compute=True, write_metadata_file=True)
                self.cluster_manager.client.cancel(view)
            if persist:
                return dd.read_parquet(f"{view_path}").persist()
            return dd.read_parquet(f"{view_path}")
        return fallback()

    def save_bottlenecks(self, bottlenecks: dd.DataFrame, partition_size='64MB'):
        bottleneck_dir = self.bottleneck_dir
        if not bottleneck_dir:
            bottleneck_dir = f"{self.working_dir}/{self.run_id}/bottlenecks"
        bottlenecks \
            .repartition(partition_size=partition_size) \
            .to_parquet(bottleneck_dir, compute=True, write_metadata_file=True)
        return bottleneck_dir

    @staticmethod
    def store_extra_data(data: Tuple[Dict], data_path: str):
        with open(data_path, 'w') as f:
            return json.dump(data[0], f, cls=NpEncoder)

    @staticmethod
    def view_permutations(view_types: List[ViewType]):
        def _iter_permutations(r: int):
            return it.permutations(view_types, r + 1)
        return it.chain.from_iterable(map(_iter_permutations, range(len(view_types))))

    def _compute_view(
        self,
        records: dd.DataFrame,
        view_type: str,
        metric: str,
        metric_boundary: dd.core.Scalar,
    ) -> dd.DataFrame:
        view_types = records.index._meta.names

        non_proc_agg_dict = self._get_agg_dict(
            for_view_type=view_type,
            view_columns=records.columns,
            view_types=view_types,
            is_proc=False,
        )
        proc_agg_dict = self._get_agg_dict(
            for_view_type=view_type,
            view_columns=records.columns,
            view_types=view_types,
            is_proc=True,
        )

        # Check view type
        if view_type is not COL_PROC_NAME and COL_PROC_NAME in view_types:
            view = records \
                .reset_index() \
                .groupby([view_type, COL_PROC_NAME]) \
                .agg(non_proc_agg_dict) \
                .groupby([view_type]) \
                .agg(proc_agg_dict)
        else:
            view = records \
                .reset_index() \
                .groupby([view_type]) \
                .agg(non_proc_agg_dict)

        # Set metric slope
        view = view.map_partitions(set_metric_slope, metric=metric)

        # Return view
        return view

    @staticmethod
    def _get_agg_dict(
        for_view_type: ViewType,
        view_columns: List[str],
        view_types: List[ViewType],
        is_proc=False,
    ):
        if is_proc:
            agg_dict = {col: max if 'time' in col else sum for col in view_columns}
        else:
            agg_dict = {col: sum for col in view_columns}

        # agg_dict['bw'] = max
        # agg_dict['intensity'] = max
        # agg_dict['iops'] = max
        agg_dict['size_min'] = min
        agg_dict['size_max'] = max

        unwanted_agg_cols = ['id', for_view_type]
        for agg_col in unwanted_agg_cols:
            if agg_col in agg_dict:
                agg_dict.pop(agg_col)

        return agg_dict

    def _set_derived_columns(self, ddf: dd.DataFrame):
        # Derive `io_cat` columns
        for col in ['time', 'size', 'count']:
            for io_cat in list(IOCategory):
                col_name = f"{io_cat.name.lower()}_{col}"
                ddf[col_name] = 0.0 if col == 'time' else 0
                ddf[col_name] = ddf[col_name].mask(ddf['io_cat'] == io_cat.value, ddf[col])
        for io_cat in list(IOCategory):
            min_name, max_name = f"{io_cat.name.lower()}_min", f"{io_cat.name.lower()}_max"
            ddf[min_name] = 0
            ddf[max_name] = 0
            ddf[min_name] = ddf[min_name].mask(ddf['io_cat'] == io_cat.value, ddf['size_min'])
            ddf[max_name] = ddf[max_name].mask(ddf['io_cat'] == io_cat.value, ddf['size_max'])
        # Derive `data` columns
        ddf['data_count'] = ddf['write_count'] + ddf['read_count']
        ddf['data_size'] = ddf['write_size'] + ddf['read_size']
        ddf['data_time'] = ddf['write_time'] + ddf['read_time']
        # Derive `acc_pat` columns
        for col_suffix, col_value in zip(ACC_PAT_SUFFIXES, ['data_time', 'data_size', 'data_count']):
            for acc_pat in list(AccessPattern):
                col_name = f"{acc_pat.name.lower()}_{col_suffix}"
                ddf[col_name] = 0.0 if col_suffix == 'time' else 0
                ddf[col_name] = ddf[col_name].mask(ddf['acc_pat'] == acc_pat.value, ddf[col_value])
        # Derive metadata operation columns
        for col in ['time', 'count']:
            for md_op in DERIVED_MD_OPS:
                col_name = f"{md_op}_{col}"
                ddf[col_name] = 0.0 if col == 'time' else 0
                if md_op in ['close', 'open']:
                    ddf[col_name] = ddf[col_name].mask(ddf['func_id'].str.contains(
                        md_op) & ~ddf['func_id'].str.contains('dir'), ddf[col])
                else:
                    ddf[col_name] = ddf[col_name].mask(ddf['func_id'].str.contains(md_op), ddf[col])
        # Return ddf
        return ddf

    def _set_logical_columns(self, view: dd.DataFrame, view_types: List[ViewType]) -> dd.DataFrame:
        # Check if view types include `proc_name`
        if COL_PROC_NAME in view_types:
            view = view.map_partitions(set_proc_name_parts)

        # Check if view types include `file_name`
        if COL_FILE_NAME in view_types:
            view = view \
                .map_partitions(set_file_dir) \
                .map_partitions(set_file_pattern)

        return view

    @staticmethod
    def _wait_all(tasks: Union[dd.DataFrame, Delayed, dict]):
        if WAIT_ENABLED:
            if isinstance(tasks, dd.DataFrame):
                _ = wait(tasks)
            else:
                all_tasks, _ = unpack_collections(tasks)
                _ = wait(all_tasks)



# import subprocess
# from os import path
from pathlib import Path
from setuptools import find_packages, setup
# from skbuild import setup
# from skbuild.command.build import build
# from skbuild.command.install import install
# from skbuild.command.egg_info import egg_info


# def update_submodules():
#     print("Updating submodules...")
#     if path.exists('.gitmodules'):
#         subprocess.check_call(['git', 'submodule', 'init'])
#         subprocess.check_call(['git', 'submodule', 'update'])


# class build_with_submodules(build):
#     def run(self):
#         update_submodules()
#         build.run(self)


# class egg_info_with_submodules(egg_info):
#     def run(self):
#         update_submodules()
#         egg_info.run(self)


# class install_with_submodules(install):
#     def run(self):
#         update_submodules()
#         install.run(self)


setup(
    name="wisio",
    version="0.0.1",
    url="https://github.com/izzet/wisio",
    author="Izzet Yildirim",
    author_email="izzetcyildirim@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8, <4",
    # cmake_source_dir="tools",
    # cmdclass={
    #     "build": build_with_submodules,
    #     "egg_info": egg_info_with_submodules,
    #     "install": install_with_submodules,
    # },
    entry_points={
        "console_scripts": [
            "wisio=wisio.__main__:main",
        ]
    },
    packages=find_packages(),
    install_requires=[
        "dask>=2023.9.0",
        "dask_jobqueue==0.8.2",
        "distributed>=2023.9.0",
        "inflect==7.0",
        "jinja2>=3.0",
        "numpy==1.24.3",
        "matplotlib==3.2.1",
        "pandas>=2.1",
        "pyyaml>=5.4",
        "rich==13.6.0",
        "scikit-learn>=1.4",
        "scipy>=1.10",
        "venn==0.1.3",
    ],
    extras_require={"darshan": ["darshan>=3.4"]},
)



import dask.dataframe as dd
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Literal, Optional, Union, Tuple
from .constants import HUMANIZED_VIEW_TYPES


class Score(Enum):
    NONE = 'none'
    TRIVIAL = 'trivial'
    VERY_LOW = 'very low'
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'
    VERY_HIGH = 'very high'
    CRITICAL = 'critical'


AnalysisAccuracy = Literal['accurate', 'optimistic', 'pessimistic']
Metric = Literal[
    'att_perf',
    'bw',
    'intensity',
    'iops',
    'time',
]
OutputType = Literal['console', 'csv', 'html', 'json', 'sqlite']
ViewType = Literal['file_name', 'proc_name', 'time_range']
ViewKey = Union[Tuple[ViewType], Tuple[ViewType, ViewType],
                Tuple[ViewType, ViewType, ViewType]]


@dataclass
class AnalysisRuntimeConfig:
    accuracy: AnalysisAccuracy
    checkpoint: bool
    cluster_type: str
    debug: bool
    memory: int
    num_threads_per_worker: int
    num_workers: int
    processes: bool
    run_id: str
    slope_threshold: float
    verbose: bool
    working_dir: str


@dataclass
class ScoringResult:
    critical_view: dd.DataFrame
    records_index: dd.Index
    scored_view: dd.DataFrame


@dataclass
class RawStats:
    job_time: dd.core.Scalar
    time_granularity: int
    total_count: dd.core.Scalar


@dataclass
class RuleReason:
    condition: str
    message: str


@dataclass
class Rule:
    name: str
    condition: str
    reasons: Optional[List[RuleReason]]
    # source: Optional[str]


@dataclass
class RuleResultReason:
    description: str
    # value: Optional[float]


@dataclass
class RuleResult:
    compact_desc: Optional[str]
    description: str
    detail_list: Optional[List[str]]
    extra_data: Optional[dict]
    object_hash: Optional[int]
    reasons: Optional[List[RuleResultReason]]
    value: Optional[Union[float, int, tuple]]
    value_fmt: Optional[str]


@dataclass
class BottleneckOutput:
    description: str
    id: int
    metric: str
    num_files: int
    num_processes: int
    num_time_periods: int
    object_hash: int
    reasons: List[RuleResultReason]
    rule: str
    score: str
    view_name: str


@dataclass
class BottleneckResult:
    results: List[RuleResult]
    severities: Dict[str, int]


@dataclass
class ViewResult:
    critical_view: dd.DataFrame
    metric: str
    records: dd.DataFrame
    view: dd.DataFrame
    view_type: ViewType


MainView = dd.DataFrame


Characteristics = Dict[str, RuleResult]

ScoringPerView = Dict[ViewKey, ScoringResult]
ScoringPerViewPerMetric = Dict[Metric, ScoringPerView]

ViewResultsPerView = Dict[ViewKey, ViewResult]
ViewResultsPerViewPerMetric = Dict[Metric, ViewResultsPerView]


def humanized_view_name(view_key_type: Union[ViewKey, ViewType], separator='_'):
    if isinstance(view_key_type, tuple):
        return separator.join([HUMANIZED_VIEW_TYPES[view_type] for view_type in view_key_type])
    return HUMANIZED_VIEW_TYPES[view_key_type]


def view_name(view_key_type: Union[ViewKey, ViewType], separator='_'):
    return separator.join(view_key_type) if isinstance(view_key_type, tuple) else view_key_type
